import json
from dataclasses import dataclass

from pytest import fixture

from melvin.tecton_gen_ai.api import Agent
from melvin.tecton_gen_ai.testing import make_local_batch_feature_view


@dataclass
class OpenAIConfig:
    model: str
    seed: int
    temperature: float
    response_format: dict

    def to_langchain_kwargs(self):
        return {
            "model": self.model,
            "seed": self.seed,
            "temperature": self.temperature,
            "model_kwargs": {"response_format": self.response_format},
        }

    def to_llamaindex_kwargs(self):
        return {
            "model": self.model,
            "additional_kwargs": {"response_format": self.response_format},
            "temperature": self.temperature,
        }


@fixture
def model_config():
    return OpenAIConfig(
        model="gpt-4o-2024-08-06",
        seed=0,
        temperature=0,
        response_format={"type": "json_object"},
    )


@fixture
def langchain_llm(model_config):
    import langchain
    from langchain_openai import ChatOpenAI

    from melvin.tecton_gen_ai.utils.config_wrapper import as_config

    langchain.debug = True

    Conf = as_config(ChatOpenAI)

    return Conf(**model_config.to_langchain_kwargs())


@fixture
def llamaindex_llm(model_config):
    from llama_index.llms.openai import OpenAI

    from melvin.tecton_gen_ai.utils.config_wrapper import as_config

    Conf = as_config(OpenAI)

    return Conf(**model_config.to_llamaindex_kwargs())


@fixture
def mock_agent_service(tecton_unit_test, mock_knowledge):
    user_info = make_local_batch_feature_view(
        "user_info",
        {"user_id": "user2", "name": "Jim"},
        ["user_id"],
        description="Getting user name",
    )

    def sys_prompt():
        return (
            # "You are serving user whose user_id " + user_id + ". "
            "The result should be in json format, the key is always 'result'"
        )

    def get_tecton_employee_count() -> int:
        """
        Returns the number of employees in Tecton
        """
        return 110

    def get_tecton_female_employee_count() -> int:
        """
        Returns the number of female employees in Tecton
        """
        return 60

    return lambda llm: Agent(
        name="test",
        tools=[
            get_tecton_employee_count,
            get_tecton_female_employee_count,
            user_info,
        ],
        prompt=sys_prompt,
        knowledge=[mock_knowledge],
        llm=llm,
    )


def _test_langchain_agent(mock_agent_service, langchain_llm):
    client = mock_agent_service(langchain_llm)
    res = client.invoke(
        "how many employees in tecton that are not female",
        context={"user_id": "user2"},
    )
    assert json.loads(res)["result"] == 50

    agent = client.make_local_agent(langchain_llm.instantiate())
    with client.set_context({"user_id": "user2"}):
        res = agent.invoke({"input": "what is my name"})
    assert "jim" in json.loads(res)["result"].lower()

    with client.set_context({"user_id": "user2", "zip": "10065"}):
        res = agent.invoke({"input": "what food are for sale"})
    assert "apple" in res.lower()


def test_llamaindex_agent(mock_agent_service, llamaindex_llm):
    client = mock_agent_service(llamaindex_llm)
    res = client.invoke(
        "how many employees in tecton that are not female",
        context={"user_id": "user2"},
    )
    assert json.loads(res)["result"] == 50

    agent = client.make_local_agent(llamaindex_llm.instantiate())
    with client.set_context({"user_id": "user2"}):
        res = str(agent.chat("what is my name"))
    assert "jim" in json.loads(res)["result"].lower()

    with client.set_context({"user_id": "user2", "zip": "10065"}):
        res = str(agent.chat("what food are for sale"))
    assert "apple" in res.lower()



import os

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
LOGO_PATH = os.path.join(FILE_DIR, "imgs", "tecton.png")
MELVIN_PATH = os.path.join(FILE_DIR, "imgs", "melvin.png")
STARTING_DIR = os.getcwd()



from typing import List, Dict
from utils import _err, _log
from tecton._internals import metadata_service
from tecton import conf
from tecton_core.fco_container import FcoContainer
from tecton_proto.metadataservice.metadata_service__client_pb2 import (
    GetAllFeatureServicesRequest,
    ListWorkspacesRequest,
    QueryFeatureViewsRequest,
    GetAllTransformationsRequest,
    GetTransformationRequest,
    GetAllVirtualDataSourcesRequest,
    GetAllEntitiesRequest
)


def list_workspaces() -> List[str]:
    """
    List all workspaces in the currently connected Tecton cluster.

    Returns:
        List[str]: A list of workspace names in the connected Tecton cluster.
        Returns an empty list if the operation fails.
    """
    try:
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)
        return [ws.name for ws in response.workspaces]
    except Exception as e:
        _err(f"Failed to list workspaces: {str(e)}")
        return []


def list_feature_views() -> List[Dict]:
    """
    List all feature views in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each feature view
    """
    try:
        feature_views = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = QueryFeatureViewsRequest(workspace=workspace)
            response = metadata_service.instance().QueryFeatureViews(request)

            for fco in response.fco_container.fcos:
                if hasattr(fco, "feature_view") and fco.feature_view.fco_metadata.name:
                    feature_views.append(
                        {
                            "name": fco.feature_view.fco_metadata.name,
                            "description": fco.feature_view.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(fco.feature_view.fco_metadata.tags),
                        }
                    )

        return sorted(feature_views, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list feature views: {str(e)}")
        return []


def list_feature_services() -> List[Dict]:
    """
    List all feature services in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each feature service
    """
    try:
        feature_services = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllFeatureServicesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllFeatureServices(request)

            for feature_service in response.feature_services:
                if feature_service.fco_metadata.name:
                    feature_services.append(
                        {
                            "name": feature_service.fco_metadata.name,
                            "description": feature_service.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(feature_service.fco_metadata.tags),
                        }
                    )

        _log("Inspecting FeatureServices")

        return sorted(feature_services, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list feature services: {str(e)}")
        return []


def list_transformations() -> List[Dict]:
    """
    List all transformations in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each transformation
    """
    try:
        transformations = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllTransformationsRequest(workspace=workspace)
            response = metadata_service.instance().GetAllTransformations(request)

            for transformation in response.transformations:
                if transformation.fco_metadata.name:
                    transformations.append(
                        {
                            "name": transformation.fco_metadata.name,
                            "description": transformation.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(transformation.fco_metadata.tags),
                        }
                    )

        return sorted(transformations, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list transformations: {str(e)}")
        return []

def get_feature_service_configuration(name: str, workspace: str) -> str:
    """
    Get the entire definition of a feature service from the Tecton cluster.

    Args:
        name: The name of the feature service to retrieve
        workspace: The workspace containing the feature service

    Returns:
        str: A serialized protobuf for a Tecton Feature Service. Also includes all objects this FS depends on, like their FeatureViews etc.

    Raises:
        Exception: If the feature service is not found or other errors occur
    """
    try:

        _log(f"Getting feature service configuration for {name} in workspace {workspace}")

        import tecton
        fs = tecton.get_feature_service(name=name, workspace=workspace)

        fs_spec = str(fs._feature_set_config)
        spec_str = f"""
        Feature Service Proto:

        {fs_spec}

        """

        return spec_str
    except Exception as e:
        _err(f"Failed to get feature view definition: {str(e)}")
        return ""


def get_feature_view_configuration(name: str, workspace: str) -> str:
    """
    Get the entire definition of a feature view from the Tecton cluster.

    Args:
        name: The name of the feature view to retrieve
        workspace: The workspace containing the feature view

    Returns:
        str: A serialized protobuf for a Tecton Feature View, defining metadata, schema, pipeline, and storage. Also includes all objects this FV depends on.

    Raises:
        Exception: If the feature view is not found or other errors occur
    """
    try:

        _log(f"Getting feature view configuration for {name} in workspace {workspace}")

        import tecton
        fv = tecton.get_feature_view(name=name, workspace=workspace)


        fv_spec = str(fv._feature_definition) # str(fv._spec)
        dependent_specs = '\n'.join(str(k) for k in fv._get_dependent_specs())
        spec_str = f"""
        Feature View Proto:

        {fv_spec}

        Proto of all objects this FV depends on (you will see references in the Feature View Proto above to these objects):

        {dependent_specs}

        """

        return spec_str
    except Exception as e:
        _err(f"Failed to get feature view definition: {str(e)}")
        return ""


def get_feature_view_code(name: str, workspace: str) -> str:
    """
    Get the code definition of a feature view from the Tecton cluster.
    This essentially gets the transformation using the feature view name and assumes its the same name as transformation name which works in most cases.

    Args:
        name: The name of the feature view to retrieve
        workspace: The workspace containing the feature view

    Returns:
        str: The Python code definition of the feature view

    Raises:
        Exception: If the feature view is not found or other errors occur
    """
    try:
        request = GetTransformationRequest(
            name=name,
            workspace=workspace,
            run_object_version_check=not conf.get_bool("SKIP_OBJECT_VERSION_CHECK"),
        )
        response = metadata_service.instance().GetTransformation(request)
        fco_container = FcoContainer.from_proto(response.fco_container)
        transformation_spec = fco_container.get_single_root()

        if transformation_spec is None:
            msg = f"Transformation '{name}' not found in workspace '{workspace}'. Try running `list_feature_views()` to view all registered feature views."
            raise ValueError(msg)

        # Validate that all required fields are present
        if not hasattr(transformation_spec, "validation_args"):
            raise ValueError("Missing validation_args in transformation_spec")
        if not hasattr(transformation_spec.validation_args, "transformation"):
            raise ValueError("Missing transformation in validation_args")
        if not hasattr(transformation_spec.validation_args.transformation, "args"):
            raise ValueError("Missing args in transformation")
        if not hasattr(
            transformation_spec.validation_args.transformation.args, "user_function"
        ):
            raise ValueError("Missing user_function in args")
        if not hasattr(
            transformation_spec.validation_args.transformation.args.user_function,
            "body",
        ):
            raise ValueError("Missing body in user_function")

        return (
            transformation_spec.validation_args.transformation.args.user_function.body
        )
    except Exception as e:
        _err(f"Failed to get feature view code: {str(e)}")
        return ""


def list_data_sources() -> List[Dict]:
    """
    List all data sources in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each data source
    """
    try:
        data_sources = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllVirtualDataSourcesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllVirtualDataSources(request)

            for data_source in response.virtual_data_sources:
                if data_source.fco_metadata.name:
                    data_sources.append(
                        {
                            "name": data_source.fco_metadata.name,
                            "description": data_source.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(data_source.fco_metadata.tags),
                        }
                    )

        return sorted(data_sources, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list data sources: {str(e)}")
        return []


def list_entities() -> List[Dict]:
    """
    List all entities in the currently connected tecton cluster

    Returns:
        List[dict]: A list of dictionaries containing name, description, workspace, and tags for each entity
    """
    try:
        entities = []
        workspaces = list_workspaces()

        for workspace in workspaces:
            request = GetAllEntitiesRequest(workspace=workspace)
            response = metadata_service.instance().GetAllEntities(request)

            for entity in response.entities:
                if entity.fco_metadata.name:
                    entities.append(
                        {
                            "name": entity.fco_metadata.name,
                            "description": entity.fco_metadata.description,
                            "workspace": workspace,
                            "tags": dict(entity.fco_metadata.tags),
                        }
                    )

        return sorted(entities, key=lambda x: x["name"])
    except Exception as e:
        _err(f"Failed to list entities: {str(e)}")
        return []



from typing import Any, List

from langchain_community.vectorstores.lancedb import LanceDB
from langchain_core.documents import Document

from .langchain import unified_langchain_vector_search


@unified_langchain_vector_search.register(LanceDB)
def _lancedb_vector_search(
    vdb: LanceDB,
    query: str,
    top_k: int,
    search_type: str = "similarity",
    **params: Any,
) -> List[Document]:
    if "filter" in params:
        expr = " AND ".join([_kv_to_sql(k, v) for k, v in params.pop("filter").items()])
        if expr != "":
            params["filter"] = expr
            params["prefilter"] = params.get("prefilter", True)
    return vdb.search(query, search_type=search_type, k=top_k, **params)


def _kv_to_sql(key: str, value: Any) -> str:
    key = "metadata." + key
    if isinstance(value, str):
        return f"{key} = '{value}'"
    if isinstance(value, bool):
        return f"{key}" if value else f"NOT {key}"
    if isinstance(value, (int, float)):
        return f"{key} = {value}"
    raise ValueError(f"Unsupported value type {type(value)}")



from pytest import fixture

from melvin.tecton_gen_ai.api import Configs, RuntimeVar, VectorDBConfig, source_as_knowledge
from melvin.tecton_gen_ai.tecton_utils._tecton_utils import set_conf
from melvin.tecton_gen_ai.testing import make_local_source, set_dev_mode
from melvin.tecton_gen_ai.testing.utils import make_local_vector_db_config
from pydantic import BaseModel, Field


class Filter(BaseModel):
    zip: int = Field(description="Zip code")


@fixture
def tecton_unit_test():
    set_dev_mode()

    with set_conf(
        {
            "TECTON_FORCE_FUNCTION_SERIALIZATION": "false",
            "DUCKDB_EXTENSION_REPO": "",
            "TECTON_SKIP_OBJECT_VALIDATION": "true",
            "TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE": "rift",
            "TECTON_BATCH_COMPUTE_MODE": "rift",
        }
    ):
        yield


@fixture
def tecton_vector_db_test_config(tmp_path):
    path = str(tmp_path / "test.db")
    return make_local_vector_db_config(path, remove_if_exists=True)


@fixture
def mock_knowledge(tecton_unit_test, tecton_vector_db_test_config):
    data = [
        {"zip": 98005, "item_id": 1, "description": "pencil"},
        {"zip": 98005, "item_id": 2, "description": "car"},
        {"zip": 98005, "item_id": 3, "description": "paper"},
        {"zip": 10065, "item_id": 4, "description": "boat"},
        {"zip": 10065, "item_id": 5, "description": "cheese"},
        {"zip": 10065, "item_id": 6, "description": "apple"},
    ]
    knowledge_src = make_local_source(
        "knowledge", data, auto_timestamp=True, description="Items for sale"
    )

    return source_as_knowledge(
        knowledge_src,
        VectorDBConfig(
            name=knowledge_src.name,
            vector_db=tecton_vector_db_test_config,
            filter=Filter,
        ),
        vectorize_column="description",
    )


@fixture
def set_default_llm(tecton_unit_test):
    with Configs(
        llm={
            "model": "openai/gpt-4o-2024-11-20",
            "api_key": RuntimeVar(name="OPENAI_API_KEY"),
            # "model": "anthropic/claude-3-5-sonnet-20241022",
            # "api_key": RuntimeVar(name="ANTHROPIC_API_KEY"),
            "temperature": 0,
        }
    ).update_default():
        yield



import asyncio

import pytest

from melvin.tecton_gen_ai.utils._coroutine import _CoroutineThread, run


async def sample_coroutine(value, delay=0):
    await asyncio.sleep(delay)
    return value


async def failing_coroutine():
    raise ValueError("Intentional error")


def test_coroutine_thread_success():
    """Test that _CoroutineThread successfully runs a coroutine."""
    coroutine = sample_coroutine(42)
    thread = _CoroutineThread(coroutine)
    thread.start()
    thread.join()
    assert thread.result == 42
    assert thread.exception is None


def test_coroutine_thread_exception():
    """Test that _CoroutineThread captures exceptions from a coroutine."""
    coroutine = failing_coroutine()
    thread = _CoroutineThread(coroutine)
    thread.start()
    thread.join()
    assert thread.exception is not None
    assert isinstance(thread.exception, ValueError)
    assert str(thread.exception) == "Intentional error"


def test_coroutine_thread_cancel():
    """Test that _CoroutineThread can cancel a running coroutine."""
    # TODO: @brian: This test is not working as expected, please fix
    return
    coroutine = sample_coroutine(42, delay=2)
    thread = _CoroutineThread(coroutine)
    thread.start()
    thread.cancel()
    thread.join()
    with pytest.raises(ValueError, match="Result not set"):
        _ = thread.result


def test_run_function_no_event_loop():
    """Test the run function when no event loop is running."""
    result = run(sample_coroutine(42))
    assert result == 42


async def test_run_function_with_event_loop():
    """Test the run function when an event loop is already running."""

    async def test_coroutine():
        return run(sample_coroutine(42))

    result = await test_coroutine()
    assert result == 42



from functools import singledispatch
from typing import Any

from ..utils.config_wrapper import ConfigWrapper
from ..utils.runtime import RuntimeVar

from .llm import _load_dependencies


def make_vector_db(obj: Any, conf: bool) -> Any:
    """
    Make a vector database object or the config wrapper. This is not for users to call directly.

    Args:

        obj: The object
        conf: whether to make a config wrapper
    """
    _load_dependencies()
    if conf:
        return make_vector_db_conf(obj)
    return make_vector_db_instance(RuntimeVar.materialize(obj))


@singledispatch
def make_vector_db_instance(obj: Any) -> Any:
    """
    Make a vector database object. This is not for users to call directly.

    Args:

        obj: The object
    """
    if isinstance(obj, ConfigWrapper):
        return make_vector_db(obj.instantiate())
    raise NotImplementedError(f"Unsupported type {type(obj)}")


@singledispatch
def make_vector_db_conf(obj: Any) -> ConfigWrapper:
    """
    Make a config wrapper of a vector database object. This is not for users to call directly.

    Args:

        obj: The object
    """
    if isinstance(obj, ConfigWrapper):
        return obj
    raise NotImplementedError(f"Unsupported type {type(obj)}")



import enum
import inspect
from typing import Any, Union, get_type_hints

from tecton.framework import base_tecton_object

from melvin.tecton_gen_ai.testing.examples.copilot import _FuncOrClass

_FCO_NAME_MAP = {
    "BatchFeatureView": "batch_feature_view",
    "StreamFeatureView": "stream_feature_view",
    "RealtimeFeatureView": "realtime_feature_view",
    "PandasBatchConfig": "pandas_batch_config",
    "PyArrowBatchConfig": "pyarrow_batch_config",
    "SparkBatchConfig": "spark_batch_config",
    "SparkStreamConfig": "spark_stream_config",
}


class APIGraphBuilder:
    def __init__(self):
        objects = list(_FuncOrClass.from_expressions(["tecton"]))
        scope = {x.name: x for x in objects}
        deps = self._find_tecton_dependencies(scope)
        res = {}
        for key, value in scope.items():
            res[key] = {"declaration": value.callable_declaration, "deps": deps[key]}
        self.graph = res

    def build_code(self, names: list[str]) -> str:
        code = {}
        for name in names:
            self._build_code(name, code)
            self._build_code(_FCO_NAME_MAP.get(name, name), code)
        return "\n\n".join(code.values())

    def _build_code(self, name: str, code: dict[str, Any]):
        if name in code:
            return
        for dep in self.graph[name]["deps"]:
            self._build_code(dep, code)
        code[name] = self.graph[name]["declaration"]

    def _find_tecton_annotations(self, cls):
        try:
            if inspect.isclass(cls):
                annotations = get_type_hints(cls.__init__)
            elif inspect.isfunction(cls):
                annotations = get_type_hints(cls)
            else:
                raise ValueError("cls must be a class or a function")
            for name, param in annotations.items():
                if is_tecton_type(param):
                    yield param
                # else check if param is union, check if any of the union is a tecton type
                elif hasattr(param, "__origin__") and param.__origin__ == Union:
                    for arg in param.__args__:
                        if is_tecton_type(arg):
                            yield arg
        except Exception:
            return None

    def _find_tecton_mentions(self, obj):
        # check obj.obj is a class type use inspect
        if inspect.isclass(obj):
            if not issubclass(obj, enum.Enum):
                yield from self._find_tecton_annotations(obj)
        elif inspect.isfunction(obj):
            yield from self._find_tecton_annotations(obj)

    def _find_tecton_dependencies(self, scope):
        res = {}
        for key, obj in scope.items():
            res[key] = list(
                x.__name__
                for x in self._find_tecton_mentions(obj.obj)
                if x.__name__ in scope
            )
        return res


def is_tecton_type(tp):
    try:
        tp_str = tp.__module__
        return tp_str.startswith("tecton.") or tp_str == "tecton"
    except Exception:
        return False


def is_tecton_object(x):
    try:
        return isinstance(x, base_tecton_object.BaseTectonObject)
    except Exception:
        return False


def extract_tecton_objects(path: str):
    with open(path, "r") as f:
        code = f.read()
    return dict(_parse(code))


def _parse(code):
    _keys = set(locals().keys())
    exec(code)
    _keys2 = set(locals().keys())
    for x in _keys2 - _keys:
        if x in ["_keys", "code"]:
            continue
        # check x is a type
        v = locals()[x]
        if is_tecton_object(v):
            yield x, v



# This file contains lru_cached functions that are used to cache the results of function calls.
# They should be replaced when the resource implementation is fixed.

import inspect
from functools import lru_cache
from typing import Callable


@lru_cache
def code_to_func(code: str) -> Callable:
    exec(code.strip())
    for v in locals().values():
        if inspect.isfunction(v):
            return v
    raise ValueError("No function found")



from ..constants import _TECTON_CHECKPOINT_ATTR, UUID_ATTR


import json
import uuid
from datetime import date, datetime
from typing import Any


def to_uuid(obj: Any) -> str:
    def _to_serializable(obj):
        if obj is None:
            return float("nan")
        elif isinstance(obj, (str, int, float, bool)):
            return obj
        elif isinstance(obj, (set, tuple, list)):
            return list(_to_serializable(x) for x in obj)
        elif isinstance(obj, dict):
            return {_to_serializable(k): _to_serializable(v) for k, v in obj.items()}
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif hasattr(obj, _TECTON_CHECKPOINT_ATTR):
            return getattr(obj, _TECTON_CHECKPOINT_ATTR)
        elif hasattr(obj, UUID_ATTR):
            return getattr(obj, UUID_ATTR)()
        raise NotImplementedError(f"Type {type(obj)} is not serializable")

    return str(uuid.uuid5(uuid.NAMESPACE_DNS, json.dumps(_to_serializable(obj))))



from melvin.tecton_gen_ai.api import Configs


def test_configs():
    with Configs(
        llm="gpt",
        fv_base_config={"x": 1},
        bfv_config={"y": 2},
        rtfv_config={"z": 3},
        sfv_config={"a": 4},
        feature_service_config={"b": 5},
    ).update_default():
        conf = Configs.get_default()
        assert conf.llm == "gpt"
        assert conf.get_bfv_config() == {"x": 1, "y": 2}
        assert conf.get_rtfv_config() == {"x": 1, "z": 3}
        assert conf.get_sfv_config() == {"x": 1, "a": 4}
        assert conf.feature_service_config == {"b": 5}

        with Configs(
            fv_base_config={"x": 2},
            bfv_config={"y": 3},
            feature_service_config={"b": 6},
        ).update_default():
            conf = Configs.get_default()
            assert conf.llm == "gpt"
            assert conf.get_bfv_config() == {"x": 2, "y": 3}
            assert conf.get_rtfv_config() == {"x": 2, "z": 3}
            assert conf.get_sfv_config() == {"x": 2, "a": 4}
            assert conf.feature_service_config == {"b": 6}

        conf = Configs.get_default()
        assert conf.llm == "gpt"
        assert conf.get_bfv_config() == {"x": 1, "y": 2}
        assert conf.get_rtfv_config() == {"x": 1, "z": 3}
        assert conf.get_sfv_config() == {"x": 1, "a": 4}
        assert conf.feature_service_config == {"b": 5}



from melvin.tecton_gen_ai.utils.pandas_utils import to_records
import pandas as pd
from pytest import raises

from datetime import datetime


def test_to_records():
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [4, 5, 6],
            "c": [7, 8, 9],
        }
    )
    assert to_records(df) == [
        {"a": 1, "b": 4, "c": 7},
        {"a": 2, "b": 5, "c": 8},
        {"a": 3, "b": 6, "c": 9},
    ]
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
        }
    )
    df["b"] = pd.to_datetime("2022-01-01")
    assert isinstance(to_records(df)[0]["b"], pd.Timestamp)
    assert to_records(df) == [
        {"a": 1, "b": pd.Timestamp("2022-01-01")},
        {"a": 2, "b": pd.Timestamp("2022-01-01")},
        {"a": 3, "b": pd.Timestamp("2022-01-01")},
    ]
    assert not isinstance(to_records(df, timestamp_to="py")[0]["b"], pd.Timestamp)
    assert to_records(df, timestamp_to="py") == [
        {"a": 1, "b": datetime(2022, 1, 1, 0, 0)},
        {"a": 2, "b": datetime(2022, 1, 1, 0, 0)},
        {"a": 3, "b": datetime(2022, 1, 1, 0, 0)},
    ]
    assert to_records(df, timestamp_to="str") == [
        {"a": 1, "b": "2022-01-01 00:00:00"},
        {"a": 2, "b": "2022-01-01 00:00:00"},
        {"a": 3, "b": "2022-01-01 00:00:00"},
    ]
    with raises(ValueError):
        to_records(df, timestamp_to="invalid")



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import ClassVar

import pytest
from tecton.types import Array, Bool, Field, Float64, Int64, String, Struct, Timestamp

from melvin.tecton_gen_ai.tecton_utils._schema_utils import (
    get_tecton_fields_from_json_schema,
    load_to_rich_dict,
)


@dataclass
class ParseTestCase:
    json_str: str
    expected: dict


@dataclass
class TectonFieldExtractionTestCase:
    name: str
    schema: dict
    expected_fields: list[Field]
    parse_test_cases: list[ParseTestCase]
    test_cases: ClassVar[dict[str, "TectonFieldExtractionTestCase"]] = {}

    @classmethod
    def register(
        cls,
        name: str,
        schema: dict,
        expected_fields: list[Field],
        parse_test_cases: list[ParseTestCase],
    ):
        cls.test_cases[name] = cls(name, schema, expected_fields, parse_test_cases)

    @classmethod
    def get_test_cases(cls) -> list[TectonFieldExtractionTestCase]:
        return list(cls.test_cases.values())

    @classmethod
    def get(cls, name: str) -> TectonFieldExtractionTestCase:
        if test_case := cls.test_cases.get(name):
            return test_case
        raise ValueError(f"Test case with name {name} not found")


TectonFieldExtractionTestCase.register(
    name="simple_schema",
    schema={
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "age": {"type": "integer"},
            "is_student": {"type": "boolean"},
        },
    },
    expected_fields=[
        Field(name="name", dtype=String),
        Field(name="age", dtype=Int64),
        Field(name="is_student", dtype=Bool),
    ],
    parse_test_cases=[
        ParseTestCase(
            json_str='{"name": "John", "age": 30, "is_student": false}',
            expected={"name": "John", "age": 30, "is_student": False},
        ),
    ],
)

TectonFieldExtractionTestCase.register(
    name="nested_schema",
    schema={
        "type": "object",
        "properties": {
            "person": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "birthdate": {"type": "string", "format": "date-time"},
                },
            },
            "scores": {
                "type": "array",
                "items": {"type": "number"},
            },
        },
    },
    expected_fields=[
        Field(
            name="person",
            dtype=Struct(
                [
                    Field(name="name", dtype=String),
                    Field(name="birthdate", dtype=Timestamp),
                ]
            ),
        ),
        Field(name="scores", dtype=Array(Float64)),
    ],
    parse_test_cases=[
        ParseTestCase(
            json_str='{"person": {"name": "John", "birthdate": "2023-01-01T00:00:00Z"}, "scores": [95.5, 88.0]}',
            expected={
                "person": {
                    "name": "John",
                    "birthdate": datetime(2023, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                },
                "scores": [95.5, 88.0],
            },
        ),
    ],
)

TectonFieldExtractionTestCase.register(
    name="ref_schema",
    schema={
        "definitions": {
            "person": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                },
            }
        },
        "type": "object",
        "properties": {
            "student": {"$ref": "#/definitions/person"},
        },
    },
    expected_fields=[
        Field(
            name="student",
            dtype=Struct(
                [Field(name="name", dtype=String), Field(name="age", dtype=Int64)]
            ),
        ),
    ],
    parse_test_cases=[
        ParseTestCase(
            json_str='{"student": {"name": "John", "age": 30}}',
            expected={"student": {"name": "John", "age": 30}},
        ),
    ],
)

TectonFieldExtractionTestCase.register(
    name="enum_ref_schema",
    # This tests both `enum` and the recursive reference to a ref
    schema={
        "defs": {  # Using 'defs' instead of 'definitions' (we want to test both scenarios since pydantic uses `defs`)
            "status": {"type": "string", "enum": ["active", "inactive", "pending"]},
            "employee": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "status": {"$ref": "#/defs/status"},
                },
            },
        },
        "type": "object",
        "properties": {
            "worker": {"$ref": "#/defs/employee"},
        },
    },
    expected_fields=[
        Field(
            name="worker",
            dtype=Struct(
                [
                    Field(name="name", dtype=String),
                    Field(name="status", dtype=String),
                ]
            ),
        ),
    ],
    parse_test_cases=[
        ParseTestCase(
            json_str='{"worker": {"name": "Alice", "status": "active"}}',
            expected={"worker": {"name": "Alice", "status": "active"}},
        ),
    ],
)


@pytest.mark.parametrize(
    "test_case", TectonFieldExtractionTestCase.get_test_cases(), ids=lambda x: x.name
)
def test_get_tecton_fields_from_json_schema(test_case: TectonFieldExtractionTestCase):
    fields = get_tecton_fields_from_json_schema(test_case.schema)
    assert len(fields) == len(test_case.expected_fields)
    for expected_field, field in zip(test_case.expected_fields, fields):
        assert expected_field.name == field.name
        assert expected_field.dtype == field.dtype


@pytest.mark.parametrize(
    "test_case", TectonFieldExtractionTestCase.get_test_cases(), ids=lambda x: x.name
)
def test_load_to_rich_dict(test_case: TectonFieldExtractionTestCase):
    for parse_test_case in test_case.parse_test_cases:
        result = load_to_rich_dict(parse_test_case.json_str, test_case.schema)
        assert result == parse_test_case.expected


if __name__ == "__main__":
    pytest.main()



import os
from dataclasses import dataclass
from typing import Any

from pytest import raises

from melvin.tecton_gen_ai.utils.config_wrapper import (
    ExternalArgument,
    RuntimeVar,
    as_config,
    from_json_config,
    to_json_config,
)
from melvin.tecton_gen_ai.utils.hashing import to_uuid
from melvin.tecton_gen_ai.utils.runtime import runtime_context


@dataclass
class TestObj:
    """This is a test"""

    a: int
    b: Any

    def __eq__(self, other):
        return self.a == other.a and str(self.b) == str(other.b)

    def __str__(self):
        return f"TestObj(a={self.a}, b={self.b})"


class TestObj3:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def compute(self):
        return self.a + self.b

    def __json_config_dict__(self):
        return {"a": self.a, "b": self.b}


def test_config_wrapper():
    TestObjConfig = as_config(TestObj)
    w = TestObjConfig(a=1, b="hello")
    assert w.target_type == TestObj
    assert w.instantiate() == TestObj(a=1, b="hello")
    from_json_config(w.to_json()) == TestObj(a=1, b="hello")

    w0 = TestObjConfig(a=1, b=None)
    w = TestObjConfig(a=1, b=w0)
    assert w.instantiate() == TestObj(a=1, b=TestObj(a=1, b=None))
    obj = from_json_config(to_json_config(w))
    assert obj == TestObj(a=1, b=TestObj(a=1, b=None))
    assert isinstance(obj.b, TestObj)

    w0 = TestObjConfig(a=ExternalArgument(name="x"), b=None)
    w = TestObjConfig(a=ExternalArgument(name="y"), b=w0)
    assert w.instantiate({"x": 1, "y": 2}) == TestObj(a=2, b=TestObj(a=1, b=None))
    with raises(KeyError):
        w.instantiate()
    from_json_config(w.to_json(), {"x": 1, "y": 2}) == TestObj(
        a=2, b=TestObj(a=1, b=None)
    )

    w0 = TestObjConfig(a=RuntimeVar(name="x"), b=None)
    w = TestObjConfig(a=RuntimeVar(name="y"), b=w0)
    with runtime_context({"tecton_secrets": {"x": 3, "y": 4}}):
        assert w.instantiate() == TestObj(a=4, b=TestObj(a=3, b=None))
        assert from_json_config(w.to_json()) == TestObj(a=4, b=TestObj(a=3, b=None))

    w = to_json_config({"a": 1, "b": {"c": RuntimeVar(name="x")}})
    with runtime_context({"tecton_secrets": {"x": 3}}):
        assert from_json_config(w) == {"a": 1, "b": {"c": 3}}

    w = to_json_config(TestObj3(1, RuntimeVar(name="x")))
    with runtime_context({"tecton_secrets": {"x": 3}}):
        assert from_json_config(w).compute() == 4


def test_temp():
    from langchain_community.vectorstores.lancedb import LanceDB
    from langchain_openai import OpenAIEmbeddings

    LanceDBConf = as_config(LanceDB)
    OpenAIEmbeddingsConf = as_config(OpenAIEmbeddings)

    vdb = LanceDBConf(embedding=OpenAIEmbeddingsConf(), uri="sdf")
    assert isinstance(
        from_json_config(to_json_config(vdb)).embeddings, OpenAIEmbeddings
    )


def test_context_var():
    v = RuntimeVar(name="_test")
    with raises(ValueError):
        v.get()
    v = RuntimeVar(name="_test", sources=["secrets"])
    with raises(ValueError):
        v.get()
    with runtime_context({"tecton_secrets": {"_test": "1"}}):
        assert v.get() == "1"
    v = RuntimeVar(name="_test", sources=["secrets", "env"])
    with raises(ValueError):
        v.get()
    with runtime_context({"tecton_secrets": {"_test": "1"}}):
        assert v.get() == "1"
    os.environ["_test"] = "2"
    assert v.get() == "2"
    with runtime_context({"tecton_secrets": {"_test": "1"}}):
        assert v.get() == "1"
    v = RuntimeVar(name="_test")
    os.environ["_test"] = "2"
    with runtime_context({"tecton_secrets": {"_test": "1"}}):
        assert v.get() == "2"
        assert RuntimeVar.materialize(v) == "2"
        assert RuntimeVar.materialize({"a": {"b": v}}) == {"a": {"b": "2"}}
        assert RuntimeVar.materialize(1) == 1

    os.environ["_test"] = "1"
    uid1 = to_uuid(v)
    os.environ["_test"] = "2"
    uid2 = to_uuid(v)
    uid3 = to_uuid(v)
    assert uid1 != uid2
    assert uid2 == uid3



import importlib
import json
from datetime import datetime
from functools import lru_cache
from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar

from pydantic import BaseModel
from typing_extensions import ParamSpec

from .runtime import RuntimeVar

T = TypeVar("T")
P = ParamSpec("P")

_TECTON_OBJECT_KEY = "__tecton_config_object__"


def as_config(object_type: Callable[P, T]) -> Callable[P, Any]:
    class _Wrapper(ConfigWrapper[object_type]):
        pass

    return _Wrapper


def to_json_config(obj: Any) -> str:
    if isinstance(obj, dict):
        obj = _DictConfig(**obj)
    tf = _EncodeTraverser()
    return json.dumps(tf(obj))


def from_json_config(
    json_str: str,
    external_kwargs: Optional[Dict[str, Any]] = None,
    use_cache: bool = True,
) -> Any:
    if use_cache:
        res = _from_json_config(json_str, json.dumps(external_kwargs))
    res = ConfigWrapper.from_json(json_str).instantiate(external_kwargs)
    if isinstance(res, _DictConfig):
        return res._kwargs
    return res


@lru_cache
def _from_json_config(json_str: str, kwargs: str) -> Any:
    return ConfigWrapper.from_json(json_str).instantiate(json.loads(kwargs))


class ExternalArgument(BaseModel):
    name: str


class ConfigWrapper(Generic[T]):
    def __init__(self, **kwargs: Any):
        self._kwargs = kwargs

    @property
    def target_type(self) -> Type[T]:
        return self.__class__.__orig_bases__[0].__args__[0]

    def instantiate(self, external_kwargs: Optional[Dict[str, Any]] = None) -> T:
        tf = _InitTraverser(external_kwargs)
        return self.target_type(**tf(self._kwargs))

    def to_json(self) -> str:
        return to_json_config(self)

    @staticmethod
    def from_json(json_str: str) -> "ConfigWrapper":
        tf = _DecodeTraverser()
        obj = tf(json.loads(json_str))
        if not isinstance(obj, ConfigWrapper):
            raise ValueError("Expected ObjectConfig object")
        return obj


class _DictConfig(ConfigWrapper[dict]):
    pass


class _Traverser:
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, list):
            return [self(o) for o in obj]
        if isinstance(obj, tuple):
            return tuple(self(o) for o in obj)
        if isinstance(obj, dict):
            return {self(k): self(v) for k, v in obj.items()}
        return obj


class _InitTraverser(_Traverser):
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, ConfigWrapper):
            return obj.instantiate(self._external_kwargs)
        if isinstance(obj, ExternalArgument):
            return self._external_kwargs[obj.name]
        if isinstance(obj, RuntimeVar):
            return obj.get()
        return super().__call__(obj)

    def __init__(self, external_kwargs: Optional[Dict[str, Any]]):
        self._external_kwargs = external_kwargs or {}


class _EncodeTraverser(_Traverser):
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, ExternalArgument):
            return {_TECTON_OBJECT_KEY: "arg", "name": obj.name}
        if isinstance(obj, RuntimeVar):
            return {
                _TECTON_OBJECT_KEY: "context_var",
                "name": obj.name,
                "sources": obj.sources,
            }
        if hasattr(obj, "__json_config_dict__"):
            obj = as_config(obj.__class__)(**obj.__json_config_dict__())
        if isinstance(obj, ConfigWrapper):
            data = {_TECTON_OBJECT_KEY: "conf"}
            data.update(self(obj._kwargs))
            data[_TECTON_OBJECT_KEY + "type"] = _type_to_str(obj.target_type)
            return data
        if isinstance(obj, datetime):
            return {_TECTON_OBJECT_KEY: "ts", "value": obj.isoformat()}
        return super().__call__(obj)


class _DecodeTraverser(_Traverser):
    def __call__(self, obj: Any) -> Any:
        if isinstance(obj, dict) and _TECTON_OBJECT_KEY in obj:
            params = obj.copy()
            otype = params.pop(_TECTON_OBJECT_KEY)
            if otype == "arg":
                return ExternalArgument(name=params["name"])
            if otype == "context_var":
                return RuntimeVar(name=params["name"], sources=params["sources"])
            if otype == "conf":
                tp = _str_to_type(params.pop(_TECTON_OBJECT_KEY + "type"))
                return as_config(tp)(**self(params))
            if otype == "ts":
                return datetime.fromisoformat(params["value"])
            raise ValueError(f"Unknown object type {otype}")
        return super().__call__(obj)


def _type_to_str(tp: Type) -> str:
    return f"{tp.__module__}.{tp.__qualname__}"


def _str_to_type(s: str) -> Type:
    parts = s.split(".")
    mod = ".".join(parts[:-1])
    obj = parts[-1]
    
    # Fix for relative imports - if it's a tecton_gen_ai module, ensure it has the right prefix
    if "tecton_gen_ai" in mod:
        # Remove src. prefix if it exists (for backward compatibility)
        if mod.startswith("src."):
            mod = mod[4:]  # Remove 'src.'
        # Ensure it starts with 'melvin.' if not a test module
        if not mod.startswith("melvin.") and not mod.startswith("tests."):
            mod = f"melvin.{mod}"
        
    return getattr(importlib.import_module(mod), obj)



# CLI Command Tools
from .cli_command_tools import get_command_help, tecton_cli_execute, tecton_cli_help

# Code Management Tools
from .code_management_tools import (
    fix_syntax_issues,
    generate_flowchart_representation,
    get_source_code,
    get_source_code_lineno,
    modify_source_code,
    save_modified_code,
    validate_with_tecton_plan,
)

# Cost Tools
from .cost_tools import SnowflakeCostQueryTool

# Feature Repo Tools
from .feature_repo_tools import (
    create_feature_repo,
    find_tecton_feature_repositories,
    move_to_folder,
)

# Metadata Service Tools
from .metadata_service_tools import (
    get_feature_view_code,
    list_data_sources,
    list_entities,
    list_feature_services,
    list_feature_views,
    list_transformations,
    list_workspaces,
    get_feature_view_configuration,
    get_feature_service_configuration
)

# System Tools
from .sys_tools import generate_graph

query_cost_data = SnowflakeCostQueryTool().query_cost_data

__all__ = [
    # CLI Command Tools
    "get_command_help",
    "tecton_cli_help",
    "tecton_cli_execute",
    # Metadata Service Tools
    "list_workspaces",
    "list_feature_views",
    "list_feature_services",
    "list_transformations",
    "get_feature_view_code",
    "list_data_sources",
    "list_entities",
    "get_feature_view_configuration",
    "get_feature_service_configuration",
    # Feature Repo Tools
    "create_feature_repo",
    "move_to_folder",
    "find_tecton_feature_repositories",
    # Code Management Tools
    "get_source_code_lineno",
    "get_source_code",
    "modify_source_code",
    "fix_syntax_issues",
    "save_modified_code",
    "validate_with_tecton_plan",
    "generate_flowchart_representation",
    # Cost Tools
    "query_cost_data",
    "generate_graph",
]



import inspect
import json
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union  # noqa

import pandas as pd
from pydantic import BaseModel

from ._tool_wrappers import FeatureViewTool, FunctionTool, SearchTool
from .base import AgentInputModel
from ..constants import _DEV_MODE_AGENT_KEY
from ..tecton_utils._internal import get_local_source_attrs, is_local_source
from ..utils._serialization import to_openai_function
from ..utils.runtime import (
    get_or_create_session_path,
    get_runtime_context,
    runtime_context,
)

from ..constants import _INPUT_JSON_FIELD, _OUTPUT_JSON_FIELD
from ..factories.llm import make_llm
from ..tecton_utils._internal import _REQUEST, make_request_source, set_serialization
from ..tecton_utils.deco import _get_from_metastore, emitter, prompt
from ..utils._serialization import serialize_pydantic
from ..utils.config_utils import Configs
from ..utils.config_wrapper import from_json_config, to_json_config
from .base import AgentBase, FeatureServiceConfig, ToolWrapper

_DEFINED_SERVICES: Dict[str, "Agent"] = {}


def _get_defined_service(name: str) -> "Agent":
    return _DEFINED_SERVICES[name]


class AgentAsTool(ToolWrapper):
    subtype: str = "agent"

    def __init__(
        self,
        agent: AgentBase,
        description: Optional[str] = None,
        use_when: Optional[Callable] = None,
    ):
        description = description or agent.description
        if not description:
            raise ValueError("Agent as tool requires a description")
        self.description = description
        self.agent = agent
        self._use_when = use_when
        self._is_remote = not isinstance(agent, Agent)

    @property
    def use_when(self) -> Optional[Callable]:
        return self._use_when

    @property
    def metastore_key(self) -> str:
        return self.agent.name

    @property
    def tool_name(self) -> str:
        return "agent_tool_" + self.agent.name

    def make_metadata(self):
        func = to_openai_function(
            self.agent.input_schema,
            name=self.agent.name,
            description=self.description,
            exclude_args=["chat_history"],
        )
        res = {
            **self.make_base_metadata(),
            "agent_name": self.agent.name,
            "function": func,
            "timeout_sec": self.agent.entrypoint_timeout_sec,
        }
        if self._is_remote:
            res["agent_json"] = to_json_config(self.agent)
        return res

    @classmethod
    def invoke(cls, agent: AgentBase, name: str, kwargs: Dict[str, Any]) -> Any:
        meta = agent.metastore["tools"][name]
        ctx = (
            agent._get_context()
        )  # TODO: should we pass the context to the remote agent?
        ctx.update(kwargs)
        query = ctx.pop("query", None)
        ctx.pop("chat_history", None)  # chat_history is not used in agent as tool
        if "agent_json" not in meta:
            _agent = agent
        else:
            _agent = from_json_config(meta["agent_json"])
        return _agent._invoke_entrypoint(
            query,
            context=ctx,
            agent_name=meta["agent_name"],
            timeout_sec=meta["timeout_sec"],
        )


class Agent(AgentBase):
    """
    The Tecton agent. The instantiated agent can either be deployed to Tecton
    or used locally for testing.

    Args:

        name: The name of the agent
        description: The description of the agent, default to None. The
            description will be required when this agent is used as a tool by
            other agents, and in this case the description is the tool description.
            If the description is provided, it will always be the first part of system
            prompt when this agent instructs its own LLM.
        prompt: the prompt for this agent to instruct its own LLM. It defaults to None.
        tools: A list of tools, agents and features views can be used as tools directly
        knowledge: A list of knowledge bases
        input_schema: The input schema of the agent, default to `str`. The input schema
            must be `str` or a subclass of `AgentInputModel`. The input schema is used to
            validate the input context of the agent. It is required when this agent is used
            as a tool by other agents and its prompt and tools need parameters.
        output_schema: The output schema of the agent, default to `str`. The output schema
            must be `str` or a subclass of pydantic BaseModel. The output schema is used to
            define the structure of the output of the agent. When it is `str` the output
            of the `invoke` method will be a string. When it is a subclass of `BaseModel`,
            the output of the `invoke` method will be a python dict following the schema.
        llm: The language model to use for this agent, default to None. It can be a string, a dict or a
            ConfigWrapper of a LLM object from Langchain or LlamaIndex.
        allow_undefined_context: Whether to allow undefined context keys in the input context.
            It defaults to True. When it is set to False, the input context must have only the keys
            defined in the `input_schema`.
        entrypoint_timeout: The timeout expression for the entrypoint, default to None. The value must be
            acceptable by `pandas.to_timedelta`. If it is None, the default timeout will be used.

    Returns:

        An Agent object

    Note:

        * The system prompt of this agent consists of `description`, `context` and `prompt`.
        * It is a good practice to define the input schema if there are extra parameters to use the agent
        * It is a good practice to set `allow_undefined_context` to False
        * It is a good practice to provide a string expression or a dict as the `llm`

    Example:

        Hello world

        ```python
        from tecton_gen_ai.api import Agent

        agent = Agent(name="hello", llm="openai/gpt-4o-mini")
        agent.invoke("Hello world")
        ```

        With a simple prompt (you don't need the prompt decorator)

        ```python
        from tecton_gen_ai.api import Agent, Configs

        # Set the default language model globally,
        # so you don't need to provide it for each agent
        Configs(llm="openai/gpt-4o-mini").set_default()

        agent = Agent(
            name="hello",
            prompt="You are talking to a 4 years old child",
        )

        agent.invoke("Hello")
        ```

        With simple tools (you don't need the tool decorator)

        ```python
        from tecton_gen_ai.api import Agent

        def get_special_message() -> str:
            return "This is a special message"

        agent = Agent(
            name="hello",
            tools=[get_special_message],
            llm="openai/gpt-4o-mini",
        )

        agent.invoke("What is the special message?")
        ```

        Prompt and tools with parameters

        ```python
        from tecton_gen_ai.api import Agent

        def sys_prompt(age:int) -> str:
            return f"You are talking to a {age} years old child"

        def get_special_message(name:str) -> str:
            '''Get a special message for a person

            Args:

                name (str): The name of the person

            Returns:

                str: The special message for the person
            '''
            return f"Hello {name}, this is magic"

        agent = Agent(
            name="hello",
            prompt=sys_prompt,
            tools=[get_special_message],
            llm="openai/gpt-4o-mini",
        )

        agent.invoke("What is the special message?", context={"name":"Jim", "age":4})

        # notice the parameters of prompts are required, but the parameters of tools are optional
        # so if you don't provide the parameters for the tools, the LLM will generate the parameters
        agent.invoke("My name is Tom, what is the special message?", context={"age":4})

        # If you want to use this agent as a tool by other agents, then you must
        # define `name` and `age` in the input schema, otherwise they will not be passed to the agent
        from pydantic import BaseModel, Field
        from tecton_gen_ai.api import AgentInputModel

        # It must inherit from `AgentInputModel`
        # The field description will be very helpful to describe the agent as a tool
        class Input(AgentInputModel):
            name: str = Field(description="The name of the person")
            age: int = Field(description="The age of the person")

        agent = Agent(
            name="hello",
            prompt=sys_prompt,
            tools=[get_special_message],
            llm="openai/gpt-4o-mini",
            input_schema=Input,
        )

        agent.invoke("What is the special message?", context={"name":"Jim", "age":4})
        ```

        It is always a good practice (no matter if you want to use the agent as a tool)
        to define all parameters required by the prompt and tools in the input schema.

        Use pydantic models to control output schema

        ```python
        from pydantic import BaseModel, Field
        from tecton_gen_ai.api import Agent, AgentInputModel, Configs

        Configs(llm="openai/gpt-4o-mini", default_timeout="20s").set_default()

        class Input(AgentInputModel):
            age: int = Field(description="The age of the person")

        # The output fields should have descriptions
        # The field description will be very helpful to help the agent generate the expected output
        class Output(BaseModel):
            title: str = Field(description="Title of the story")
            story: str = Field(description="The story")

        story_teller = Agent(
            name="story_teller",
            # description is required to describe the agent as a tool
            description="Tell a story for a person given the topic (query)",
            # prompt is to instruct this agent's LLM how to respond
            prompt="You are a story teller",

            output_schema=Output,

            # this enforces the input context to only have the keys in input schema
            # it is optional, but it is recommended to set it to False
            allow_undefined_context=False,
        )

        # the output will be a python dict following the output schema
        story_teller.invoke("Tell me a story")
        ```

        To continue, now let's use agents as tools

        ```python
        # Input must be defined and provided to the agent: poem_writer
        # because `sys_prompt` uses the `age` parameter
        # and the parameter `age` will be a part of the agent as a tool

        class Input(AgentInputModel):
            age: int = Field(description="The age of the person")

        def sys_prompt(age:int) -> str:
            return f"You are talking to a {age} years old child"

        poem_writer = Agent(
            name="song_writer",
            description="Write a poem for a person given a topic (query)",
            prompt=sys_prompt,
            input_schema=Input,
        )

        agent = Agent(
            name="general",
            tools=[story_teller, poem_writer],
        )

        agent.invoke("I am 5 years old, tell me a story and then write a song")
        ```

        Let's build a chatbot helping a student to shop for items.

        ```python
        from tecton_gen_ai.testing import make_local_source, make_local_batch_feature_view, set_dev_mode
        from tecton_gen_ai.testing.utils import make_local_vector_db_config

        set_dev_mode()  # Set the dev mode to avoid tecton login

        student = make_local_batch_feature_view(
            "student",
            {"student_id": 1, "name": "Jim", "teacher": "Mr. Smith", "preference": "fruit"},
            ["student_id"],
            description="Student information including name, teacher and shopping preference",
        )

        df = [
            {"zip":"98005", "item_id":1, "description":"pencil"},
            {"zip":"98005", "item_id":2, "description":"car"},
            {"zip":"98005", "item_id":3, "description":"paper"},
            {"zip":"10065", "item_id":4, "description":"boat"},
            {"zip":"10065", "item_id":5, "description":"cheese"},
            {"zip":"10065", "item_id":6, "description":"apple"},
        ]

        src = make_local_source(
            "for_sale",
            df,
            description="Items information",  # required for source_as_knowledge
        )
        vdb_conf = make_local_vector_db_config()

        # Create a knowledge base from the source
        from tecton_gen_ai.api import prompt, source_as_knowledge

        @prompt(sources=[student])
        def sys_prompt(student) -> str:
            return "You are serving a 4 years old child "+student["name"]

        knowledge = source_as_knowledge(
            src,
            vector_db_config=vdb_conf,
            vectorize_column="description",
            filter = [("zip", str, "the zip code of the item for sale")]
        )

        # Serve the knowledge base
        from tecton_gen_ai.api import Agent

        agent = Agent(
            name="app",
            prompt=sys_prompt,
            tools=[student],  # student is a feature view but can be used as a tool directly
            knowledge=[knowledge],
            llm = {"model": "openai/gpt-4o", "temperature": 0},
        )

        # Test locally
        with agent.set_context({"zip":"98005", "student_id":1}):
            print(agent.invoke("Suggest something for me to buy"))
        with agent.set_context({"zip":"10065", "student_id":1}):
            print(agent.invoke("Suggest something for me to buy"))
        ```
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        prompt: Any = None,
        tools: Optional[List[Any]] = None,
        knowledge: Optional[List[Any]] = None,
        emitters: Optional[List[Any]] = None,
        input_schema: Union[Type[str], Type[AgentInputModel]] = str,
        output_schema: Union[Type[str], Type[BaseModel]] = str,
        llm: Any = None,
        allow_undefined_context: bool = True,
        entrypoint_timeout: Any = None,
    ):
        from tecton import FeatureView

        super().__init__(name=name)
        self._metastore: Dict[str, Any] = {}

        self.allow_undefined_context = allow_undefined_context
        self.tools: List[Any] = []
        self.prompts: List[Any] = []
        self._metastore["entrypoint_timeout_sec"] = (
            Configs.get_default().get_timeout_sec(entrypoint_timeout)
        )

        self.knowledge_bfvs = []
        llm = llm or Configs.get_default().llm
        self.llm = None if llm is None else make_llm(llm, conf=True)
        self._input_schema = AgentInputModel if input_schema is str else input_schema
        if not issubclass(self.input_schema, AgentInputModel):
            raise ValueError(
                "input_schema must be str or a subclass of AgentInputModel"
            )
        self._metastore["input_schema"] = serialize_pydantic(self.input_schema)
        self._output_schema = output_schema
        self._metastore["output_schema"] = (
            "str"
            if self.output_schema is str
            else serialize_pydantic(self.output_schema)
        )
        if description:
            self._metastore["description"] = description

        if prompt:
            self._add_prompt(prompt)
        if tools:
            for _tool in tools:
                if isinstance(_tool, ToolWrapper):
                    self._add_tool(_tool)
                elif isinstance(_tool, FeatureView):
                    self._add_tool(FeatureViewTool(_tool, queryable=False))
                elif isinstance(_tool, AgentBase):
                    self._add_tool(AgentAsTool(_tool))
                elif callable(_tool):
                    self._add_tool(FunctionTool(_tool))
                else:
                    raise ValueError(f"Invalid tool: {_tool}")
        if knowledge:
            for k in knowledge:
                self._add_knowledge(k)
        if emitters:
            for _emitter in emitters:
                self._add_emitter(_emitter)
        self.online_fvs = [
            *self.tools,
            *self.prompts,
            self._make_metastore(),
        ]
        self.entrypoint_fv = self._make_entrypoint()
        self._run(name)
        _DEFINED_SERVICES[name] = self

        self.tool_map = {tool.name: tool for tool in self.online_fvs}
        for bfv in self.knowledge_bfvs:
            source = bfv.sources[0]
            if is_local_source(source):
                attrs = get_local_source_attrs(source)
                start = attrs["start_time"]
                end = attrs["end_time"]
                bfv.run_transformation(start, end).to_pandas()
                self.logger.info("Ingested knowledge %s to vector db", bfv.name)

    def _make_entrypoint(self):
        from tecton import Attribute, realtime_feature_view
        from tecton.types import String

        request = make_request_source(input_json=str, session_path=str)
        service_name = self.name
        connection_json = to_json_config(FeatureServiceConfig(service=service_name))
        allow_undefined_context = self.allow_undefined_context
        input_plan = serialize_pydantic(self.input_schema)
        output_plan = (
            "str"
            if self.output_schema is str
            else serialize_pydantic(self.output_schema)
        )
        llm_json = None if self.llm is None else to_json_config(self.llm)
        agent_invoke_kwargs = Configs.get_default().get_agent_invoke_kwargs()

        with set_serialization():

            @realtime_feature_view(
                name=self.name,
                sources=[request],
                mode="python",
                features=[Attribute(_OUTPUT_JSON_FIELD, String)],
                context_parameter_name="context",
                **Configs.get_default().get_rtfv_config(),
            )
            def entrypoint(request_context, context) -> str:
                import json

                from ..utils.runtime import (
                    get_or_create_session_path,
                    runtime_context,
                )

                try:
                    session_path = get_or_create_session_path(
                        request_context.get("session_path")
                    )
                    with runtime_context(
                        {}, tecton_context=context, session_path=session_path
                    ):
                        if llm_json is None:
                            raise ValueError("No language model instance provided")

                        from ..utils._serialization import (
                            deserialize_pydantic,
                        )
                        from ..utils.config_wrapper import from_json_config
                        from ..utils.runtime import (
                            make_agent_client_in_rtfv,
                        )

                        input_schema = deserialize_pydantic(input_plan)
                        output_schema = (
                            str
                            if output_plan == "str"
                            else deserialize_pydantic(output_plan)
                        )

                        input = json.loads(request_context[_INPUT_JSON_FIELD])
                        if not allow_undefined_context:
                            old = input
                            input = input_schema.model_validate(old).model_dump()
                            extra = set(old.keys()) - set(input.keys())
                            if len(extra) > 0:
                                raise ValueError(
                                    f"Undefined context is not allowed: {extra}"
                                )

                        client = make_agent_client_in_rtfv(
                            service_name=service_name, connection_json=connection_json
                        )
                        query = input.pop("query")
                        chat_history = input.pop("chat_history", [])
                        local_llm = from_json_config(llm_json)
                        res = client.invoke_locally(
                            query,
                            chat_history=chat_history,
                            context=input,
                            output_schema=output_schema,
                            local_llm=local_llm,
                            **agent_invoke_kwargs,
                        )
                        response = {
                            "result": (
                                res if isinstance(res, (str, dict)) else res.dict()
                            )
                        }
                except Exception:
                    import traceback

                    error = traceback.format_exc()

                    response = {"error": error}
                return {_OUTPUT_JSON_FIELD: json.dumps(response)}

            return entrypoint

    def _add_tool(self, wrapper: ToolWrapper) -> None:
        if "tools" not in self._metastore:
            self._metastore["tools"] = {}
        self._metastore["tools"][wrapper.metastore_key] = wrapper.make_metadata()
        rtfv = wrapper.make_rtfv()
        if rtfv is not None:
            self.tools.append(rtfv)

    def _add_emitter(self, func: Any) -> None:
        if isinstance(func, FeatureServiceConfig):
            for _emitter in func.get_all_emitters():
                self._add_emitter(_emitter)
            return
        meta = dict(_get_from_metastore(func, lambda: emitter(func)))
        if meta.get("type") != "emitter":
            raise ValueError(f"Function {func} is not an emitter")
        name = meta.pop("name")
        self.tools.append(meta.pop("fv"))
        if meta["use_when"]:
            meta["use_when"] = inspect.getsource(meta["use_when"])
        else:
            meta["use_when"] = ""
        if "emitters" not in self._metastore:
            self._metastore["emitters"] = {}
        self._metastore["emitters"][name] = meta

    def _add_prompt(self, func: Any) -> None:
        with set_serialization():
            if isinstance(func, str):
                message = func

                @prompt(name=self.name + "_prompt")
                def system_prompt() -> str:
                    return message

                func = system_prompt
            meta = dict(
                _get_from_metastore(
                    func, lambda: prompt(func=func, name=self.name + "_prompt")
                )
            )
            if meta.get("type") != "prompt":
                raise ValueError(f"Function {func} is not a prompt")
            self.prompts.append(meta.pop("fv"))
            self._metastore["prompt"] = meta

    def _add_knowledge(self, funcs: Any) -> None:
        self.knowledge_bfvs.append(funcs[0])
        self._add_tool(SearchTool(funcs[1]))

    def _make_metastore(self):
        from tecton import Attribute, realtime_feature_view
        from tecton.types import String

        _metastore = self.metastore
        name = self.name + "_metastore"

        with set_serialization():

            @realtime_feature_view(
                name=name,
                sources=[_REQUEST],
                mode="python",
                features=[Attribute(_OUTPUT_JSON_FIELD, String)],
                **Configs.get_default().get_rtfv_config(),
            )
            def metastore(request_context):
                import json

                from ..constants import _OUTPUT_JSON_FIELD

                return {_OUTPUT_JSON_FIELD: json.dumps({"result": _metastore})}

            return metastore

    def _run(self, name: str) -> list:
        from tecton import FeatureService

        kwargs = Configs.get_default().feature_service_config
        fs = [
            FeatureService(
                name=name + "_" + tool.name,
                features=[tool],
                **kwargs,
            )
            for tool in self.online_fvs
        ]
        for fv in self.knowledge_bfvs:
            fs.append(
                FeatureService(
                    name=name + "_" + fv.name,
                    **kwargs,
                    features=[fv],
                    online_serving_enabled=False,
                )
            )
        fs.append(FeatureService(name=name, **kwargs, features=[self.entrypoint_fv]))
        return fs

    @property
    def metastore(self) -> Dict[str, Any]:
        return self._metastore

    @property
    def input_schema(self) -> Type[AgentInputModel]:
        return self._input_schema

    @property
    def output_schema(self) -> Union[Type[str], Type[BaseModel]]:
        return self._output_schema

    def _invoke_entrypoint(
        self,
        message: str,
        timeout_sec: float,
        chat_history: Optional[List[Tuple[str, str]]] = None,
        context: Optional[Dict[str, Any]] = None,
        agent_name: Optional[str] = None,
    ) -> Any:
        agent = self if agent_name is None else _get_defined_service(agent_name)
        now = datetime.now()
        session_path = get_or_create_session_path()
        with runtime_context({_DEV_MODE_AGENT_KEY: agent}, session_path=session_path):
            req = {
                "query": message,
                "chat_history": (
                    [list(pair) for pair in chat_history]
                    if chat_history is not None
                    else []
                ),
            }
            if context is not None:
                req.update(context)
            output_df = self.entrypoint_fv.get_features_for_events(
                pd.DataFrame(
                    [
                        {
                            _INPUT_JSON_FIELD: json.dumps(req),
                            "session_path": session_path,
                            "_dummy_request_ts": now,
                        }
                    ]
                ),
            ).to_pandas()
            resp = json.loads(output_df[self.name + "__" + _OUTPUT_JSON_FIELD].iloc[0])

            if "error" in resp:
                raise Exception(resp["error"])
            return resp["result"]

    def _get_feature_value(
        self,
        name: str,
        key_map: Dict[str, Any],
        request_map: Dict[str, Any],
        feature_type: str,
        timeout_sec: float,
    ):
        from tecton import RequestSource

        fv = self.tool_map[name]
        res = dict(key_map)
        now = datetime.now()
        # add dummy ts anyway, and if fvs have ts field, use that to replace
        ts_field = "_dummy_request_ts"
        res[ts_field] = now
        for source in fv.sources:
            if not isinstance(source, RequestSource):
                src = source.feature_definition
                new_ts_field = src.get_timestamp_field()
                if ts_field != "_dummy_request_ts" and ts_field != new_ts_field:
                    raise ValueError(
                        "All feature views must have the same timestamp field, found: %s %s",
                        ts_field,
                        new_ts_field,
                    )
                ts_field = new_ts_field
                res[ts_field] = now

        if feature_type == "prompt_not_in_use":  # pragma: no cover
            # TODO: prompt FCO: fix or delete?
            res.update(request_map)
            if len(res) == 0:
                res["dummy"] = 0
            output_df = fv.get_prompts_for_events(
                pd.DataFrame([res]), timestamp_key=ts_field
            ).to_pandas()
            return output_df[name + "__prompt"].iloc[0]
        else:
            res.update(
                {
                    "name": name,
                    "input": json.dumps(request_map),
                    "session_path": get_runtime_context().get("session_path", ""),
                }
            )
            output_df = fv.get_features_for_events(
                pd.DataFrame([res]), timestamp_key=ts_field
            ).to_pandas()
            resp = json.loads(output_df[name + "__" + _OUTPUT_JSON_FIELD].iloc[0])
            if "error" in resp:
                raise Exception(resp["error"])
            result = resp["result"]
            return result


def fv_as_tool(
    fv: Any,
    queryable: bool,
    description: Optional[str] = None,
    timeout: Any = None,
) -> "FeatureViewTool":
    """Convert a feature view to a tool

    Args:

        fv: The feature view object
        queryable: Whether the tool is queryable using its entity ids
        description: The description of the tool, default to None. If provided,
            it will be used as the description of the tool
        timeout: The timeout expression for the tool, default to None. The value must be
            acceptable by `pandas.to_timedelta`. If it is None, the default timeout will be used.

    Returns:

        A FeatureViewTool object
    """
    return FeatureViewTool(fv, queryable, description=description, timeout=timeout)



import asyncio
import base64
import os
import tempfile
import threading
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
from pyarrow.fs import FileSystem
from tecton import Attribute, BatchSource, Entity, RequestSource, pandas_batch_config
from tecton.types import Field
from tecton_core import conf as tecton_conf

from .convert import to_tecton_type

_CACHE_PREFIX = "_tecton_cache"


def make_entity(description: Optional[str] = None, **fields) -> Entity:
    """
    Make a Tecton entity

    Args:

        description: The description of the entity, defaults to None
        **fields: The fields, where the key is the field name and the value is the
            field type in type annotation format. For example, ``name=str``

    Returns:

        Entity: The entity

    Examples:

        ```python
        from ..tecton_utils import make_entity

        user = make_entity(user_id=int, description="User entity")
        ```
    """
    _fields = [Field(k, to_tecton_type(v)) for k, v in fields.items()]
    name = _fields[0].name
    return Entity(name=name, description=description, join_keys=_fields)


def make_request_source(**fields) -> RequestSource:
    """
    Make a Tecton request source

    Args:

        **fields: The fields, where the key is the field name and the value is the
            field type in type annotation format. For example, ``name=str``

    Returns:

        RequestSource: The request source

    Examples:

        ```python
        from ..tecton_utils import make_request_source

        params = make_request_source(name=str, field1=int, field2=float, field3=bool)
        ```
    """
    return RequestSource(
        schema=[Field(k, to_tecton_type(v)) for k, v in fields.items()]
    )


def get_df_schema(df: pd.DataFrame, as_attributes: bool = False) -> List[Any]:
    """
    Get the Tecton schema of the DataFrame

    Args:

        df: The DataFrame
        as_attributes: Whether to return the schema as attributes, defaults to False

    Returns:

        List[Any]: The schema of the DataFrame
    """
    res: List[Any] = []
    for field in pa.Schema.from_pandas(df):
        res.append((field.name, to_tecton_type(field.type)))
    if not as_attributes:
        return [Field(name, type) for name, type in res]
    return [Attribute(name, type) for name, type in res]


def set_secrets_env(secrets_dict: Dict[str, Any]) -> None:
    """
    Set the secrets in the environment variables

    Args:

        secrets_dict: A dictionary of secrets

    Examples:

        ```python
        from tecton import Secret

        secrets = set_secrets_env({"env_var_name": Secret(scope="", key="")})

        @batch_feature_view(
            sources=[your_source, secrets],
            ...
        )
        def your_feature_view(your_source, secrets):
            # no need to do anything with secrets, they are already set in the environment
            ...
        ```
    """

    @pandas_batch_config(
        secrets=secrets_dict,
    )
    def api_secrets(secrets):
        import os

        import pandas as pd

        for k, v in secrets.items():
            os.environ[k] = v

        return pd.DataFrame([[0]], columns=["dummy"])

    return BatchSource(name="secrets", batch_config=api_secrets)


def get_cache_base_path() -> Tuple[FileSystem, str]:
    """
    Get the base path for the cache

    Returns:

        Tuple[FileSystem, str]: The filesystem and the path
    """
    if "MATERIALIZATION_TASK_PARAMS" in os.environ:
        try:
            from tecton_proto.materialization.params_pb2 import (
                MaterializationTaskParams,
            )
        except ImportError:
            from tecton_proto.materialization.params__client_pb2 import (
                MaterializationTaskParams,
            )

        params = MaterializationTaskParams()
        params.ParseFromString(
            base64.standard_b64decode(os.environ["MATERIALIZATION_TASK_PARAMS"])
        )
        store_path = params.offline_store_path.rstrip("/")
        if not store_path.endswith("data"):
            raise ValueError(f"Invalid offline store path: {store_path}")
        path = store_path.rsplit("/", 2)[0]
    elif "TECTON_CACHE_PATH" in os.environ:
        path = os.environ["TECTON_CACHE_PATH"]
    else:
        path = tempfile.gettempdir()

    fs, p = FileSystem.from_uri(os.path.join(path, _CACHE_PREFIX))
    fs.create_dir(p, recursive=True)
    return fs, p


def run_async(coro: Any) -> Any:
    """
    Run the coroutine asynchronously in both Tecton and Jupyter notebooks

    Args:

        coro: The coroutine to run

    Returns:

        Any: The result of the coroutine
    """

    res = [None]
    exc = [None]

    def _run():
        loop = asyncio.new_event_loop()
        try:
            res[0] = loop.run_until_complete(coro)
        except Exception as e:
            exc[0] = e
        finally:
            loop.close()

    _thread = threading.Thread(target=_run)
    _thread.start()
    _thread.join()

    if exc[0] is None:
        return res[0]
    raise exc[0]


def run_async_jobs(jobs: List[Any], concurrency: int) -> List[Any]:
    """
    Run the list of coroutines asynchronously in both Tecton and Jupyter notebooks

    Args:

        jobs: The list of coroutines to run
        concurrency: The number of concurrent jobs to run

    Returns:

        List[Any]: The results of the coroutines
    """

    async def _run(job, sem):
        async with sem:
            return await job

    async def _gather():
        sem = asyncio.Semaphore(concurrency)
        _jobs = [_run(job, sem) for job in jobs]
        return await asyncio.gather(*_jobs)

    return run_async(_gather())


def featurize_async(
    df: pd.DataFrame,
    input_column: str,
    output_column: str,
    featurizer: Any,
    concurrency: int,
    dedup=True,
    mini_batch_size=0,
):
    """
    Featurize a DataFrame using asynchronous featurizer

    Args:

        df: The DataFrame to featurize
        input_column: The input column
        output_column: The output column
        featurizer: The async featurizer function
        concurrency: The number of concurrent jobs to run
        dedup: Whether to deduplicate the results, defaults to True
        mini_batch_size: The mini batch size, defaults to 0

    Returns:

        pd.DataFrame: The DataFrame with the output column

    Note:

        If the featurizer transforms single items, mini_batch_size must be 0,
        else you must specify the mini_batch_size to be a positive number
    """

    def _extract(values: List[Any]) -> List[Any]:
        if len(values) == 0:
            return []

        if dedup:
            keys = list(set(values))
            features = _extract(keys)
            _dict = dict(zip(keys, features))
            return [_dict[x] for x in values]

        if mini_batch_size == 0:
            return run_async_jobs(
                [featurizer(value) for value in values],
                concurrency=concurrency,
            )
        else:
            jobs = []
            for i in range(0, len(values), mini_batch_size):
                jobs.append(featurizer(values[i : i + mini_batch_size]))
            arrays = run_async_jobs(jobs, concurrency=concurrency)
            return [item for sublist in arrays for item in sublist]

    features = _extract(df[input_column].tolist())
    return df.assign(**{output_column: features})


@contextmanager
def set_conf(conf: dict):
    old_conf = {}
    for k, v in conf.items():
        old_conf[k] = tecton_conf.get_or_none(k)
        tecton_conf.set(k, v)
    try:
        yield
    finally:
        for k, v in old_conf.items():
            tecton_conf.set(k, v)



import logging
import re
from typing import Any, Callable, Dict, Optional

import ipywidgets as widgets
from IPython.display import HTML, Markdown, display

from ..agent.base import AgentBase
from ..testing.log_utils import (
    _WidgetLogHandler,
    get_ui_logger,
    set_ui_logger,
)

from .utils import print_md


def qna(
    client: AgentBase,
    *,
    query: Optional[str] = None,
    context: Any = None,
    debug: bool = True,
    diagram: bool = False,
) -> Any:
    def _run(message: str) -> str:
        logger = get_ui_logger()
        logger.setLevel(logging.DEBUG)
        with client.set_logger(logger):
            return client.invoke(message, context=context)

    if not query:
        return single_turn(
            _run, realtime=False, markdown=True, debug=debug, diagram=diagram
        )

    if not debug:
        res = client.invoke(query, context=context)
    else:
        debugo = widgets.Output()
        display(debugo)
        handler = _WidgetLogHandler(debugo, diagram=diagram)
        with set_ui_logger(hanlder=handler):
            res = _run(query)
    print_md(res)


def auto_complete(
    client: AgentBase,
    search_name: str,
    handle: Any,
    top_k: int = 5,
    debug: bool = False,
) -> Any:
    if isinstance(handle, str):

        def _handle(x):
            return x[handle]

    elif handle is None:

        def _handle(x):
            return str(x)

    else:
        _handle = handle
    return single_turn(
        lambda x: "\n".join(_handle(x) for x in client.search(search_name, x, top_k)),
        realtime=True,
        markdown=False,
        debug=debug,
    )


def chat(
    client: AgentBase, debug: bool = True, context: Optional[Dict[str, Any]] = None
) -> Any:
    chat = _Chat(client, context=context)
    chat.display()


def single_turn(
    on_compute: Callable[[str], str],
    realtime: bool = False,
    markdown: bool = False,
    debug: bool = True,
    diagram: bool = False,
) -> Any:
    # Create a text input widget
    text_input = widgets.Text(
        value="",
        placeholder="Type something",
        disabled=False,
        continuous_update=realtime,
        layout=widgets.Layout(
            width="90%",
            border_radius="10px",
        ),
    )

    output = widgets.Output()
    debugo = widgets.Output()

    def on_event(change):
        with output:
            if not realtime:
                output.clear_output()
                display(Markdown("Generating response..."))
        handler = _WidgetLogHandler(debugo, diagram=diagram) if debug else None
        with set_ui_logger(hanlder=handler):
            res = on_compute(change["new"])
        with output:
            output.clear_output()
            if markdown:
                display(Markdown(res))
            else:
                print(res)

    text_input.observe(on_event, names="value")

    items = [text_input, output]
    if debug:
        accordion = widgets.Accordion(children=[debugo], titles=("Debug",))
        items.append(accordion)

    vbox = widgets.VBox(items)

    # Display the text input widget
    display(vbox)


class _Chat:
    def __init__(self, client: AgentBase, context: Optional[Dict[str, Any]] = None):
        self.box = widgets.Output()
        self.input = widgets.Textarea(
            value="",
            placeholder="Chat with AI",
            rows=5,
            disabled=False,
            layout=widgets.Layout(width="90%"),
        )
        self.submit = widgets.Button(
            description="Submit", layout=widgets.Layout(width="100px")
        )
        hbox = widgets.HBox([self.submit, self.input])
        self.vbox = widgets.VBox([self.box, hbox])
        self.history = []
        self.client = client
        self.context = context

    def display(self):
        display(_CSS)
        self.submit.on_click(self.on_submit)
        display(self.vbox)

    def append(self, role: str, text: str):
        import markdown

        self.history.append((role, text))
        rs = "chat_user" if role == "user" else "chat_agent"
        _text = markdown.markdown(text, extensions=["fenced_code", "codehilite"])
        _text = re.sub("(^<P>|</P>$)", "", _text, flags=re.IGNORECASE)
        q = (
            f'<div class="chat_outer"><div class="chat_role">{role}:'
            f'</div><div class="chat_text {rs}">{_text}</div></div>'
        )
        self.box.append_display_data(HTML(q))

    def on_submit(self, change):
        question = self.input.value
        self.input.disabled = True
        self.input.value = "Generating response ..."
        self.submit.disabled = True
        try:
            self.ask(question)
        finally:
            self.input.value = ""
            self.input.disabled = False
            self.submit.disabled = False

    def ask(self, message: str):
        self.append("user", message)
        response = self.client.invoke(
            message,
            chat_history=self.history,
            context=self.context,
        )
        self.append("ai", response)
        return response


_CSS = HTML(
    """<style>
.chat_outer {
  overflow: hidden;
}

.chat_role {
  width: 100px;
  float: left;
  text-align: right;
  font-weight: bold;
  padding-right: 10px;
}

.chat_text {
  overflow: auto;
}

.chat_agent {
  background-color: #f0fff0;
}

.chat_user {
  background-color: #f0f0ff;
}
</style>"""
)



import json
from typing import Any, Callable, Dict, List, Optional

import numpy as np

from ..constants import _OUTPUT_JSON_FIELD
from ..utils._serialization import to_openai_function
from ..utils.config_utils import Configs
from .base import AgentBase, ToolWrapper


class FunctionTool(ToolWrapper):
    def __init__(self, func: Callable):
        from ..tecton_utils.deco import _get_from_metastore, tool

        meta = dict(_get_from_metastore(func, lambda: tool(func)))
        if meta.get("type") != "tool":
            raise ValueError(f"Function {func} is not a tool")
        self._name = meta.pop("name")
        self.fv = meta.pop("fv")
        self._use_when = meta.pop("use_when")
        self._meta = meta

    @property
    def use_when(self) -> Optional[Callable]:
        return self._use_when

    @property
    def metastore_key(self) -> str:
        return self._name

    @property
    def tool_name(self) -> str:
        return self._name

    def make_metadata(self):
        return {**self.make_base_metadata(), **self._meta}

    def make_rtfv(self):
        return self.fv

    @classmethod
    def invoke(cls, agent: AgentBase, name: str, kwargs: Dict[str, Any]) -> Any:
        meta = agent.metastore["tools"][name]
        ctx = agent._get_context()
        ctx.update(kwargs)
        entity_args = meta.get("entity_args", [])
        llm_args = meta.get("llm_args", [])
        return agent._invoke(
            name,
            entity_args,
            llm_args,
            ctx,
            feature_type="tool",
            timeout_sec=meta["timeout_sec"],
        )


class SearchTool(FunctionTool):
    subtype: str = "search"

    @classmethod
    def invoke(cls, agent: AgentBase, name: str, kwargs: Dict[str, Any]) -> Any:
        kwargs = kwargs.copy()
        _filters = json.loads(kwargs.pop("filter", None) or "{}")
        _fctx = agent._get_context()
        _fctx.update(_filters)
        kwargs["filter"] = json.dumps(_fctx)
        return super().invoke(agent, name, kwargs)


def _transform_group_result(fv: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Transform the group result to a list of dictionaries

    Args:

        fv: a dict of single value or arrays

    Returns:

        List[Dict[str, Any]]: a list of dictionaries
    """
    cols: List[List[Any]] = [
        v.tolist() for v in fv.values() if isinstance(v, np.ndarray)
    ]
    if len(cols) == 0:
        return fv
    elif len(cols) < len(fv):
        raise ValueError("All columns must be numpy arrays or none")
    blobs = []
    keys = list(fv.keys())
    step = len(fv) // 2
    for i in range(step):
        key_name = keys[i].rsplit("_keys_", 1)[0]
        val_name = keys[i + step]
        blob = [{key_name: x, val_name: y} for x, y in zip(cols[i], cols[i + step])]
        blobs.extend(blob)
    return blobs


class FeatureViewTool(ToolWrapper):
    subtype: str = "fv"

    def __init__(
        self,
        fv: Any,
        queryable: bool,
        description: Optional[str] = None,
        use_when: Optional[Callable] = None,
        timeout: Any = None,
    ):
        if not hasattr(fv, "entities") or len(fv.entities) != 1:
            raise ValueError(f"FeatureView {fv} must have exactly one entity")
        self.description = description or fv.description
        if not self.description:
            raise ValueError(
                f"FeatureView {fv} must have a description or you "
                "should provide the description from `fv_as_tool`"
            )

        self.fv = fv
        self.queryable = queryable
        self.timeout = timeout
        self._use_when = use_when

    @property
    def use_when(self) -> Optional[Callable]:
        return self._use_when

    @property
    def metastore_key(self) -> str:
        return self.fv.name

    @property
    def tool_name(self) -> str:
        return "fv_tool_" + self.fv.name

    def make_metadata(self):
        from ..tecton_utils.deco import _get_source_names
        from ..tecton_utils._internal import entity_to_pydantic

        fv = self.fv
        # notice transformation_schema doesn't consider aggregation, so this is not the final schema
        # gettting entity schema from it is fine
        model = entity_to_pydantic(fv.entities[0], fv.transformation_schema())
        func = to_openai_function(
            model, name=self.tool_name, description=self.description
        )
        args = list(model.model_fields.keys())
        return {
            **self.make_base_metadata(),
            "queryable": self.queryable,
            "args": args,
            "function": func,
            "source_names": _get_source_names([fv]),
            "timeout_sec": Configs.get_default().get_timeout_sec(self.timeout),
        }

    def make_rtfv(self):
        from tecton import Attribute, realtime_feature_view
        from tecton.types import String

        from ..tecton_utils._internal import _REQUEST, set_serialization

        fv = self.fv
        tool_name = self.tool_name
        with set_serialization():

            @realtime_feature_view(
                name=tool_name,
                sources=[_REQUEST, fv],
                mode="python",
                features=[Attribute(_OUTPUT_JSON_FIELD, String)],
                **Configs.get_default().get_rtfv_config(),
            )
            def fv_tool(request_context, _fv) -> str:
                import json

                from ..constants import _OUTPUT_JSON_FIELD

                if tool_name != request_context["name"]:
                    return {_OUTPUT_JSON_FIELD: "{}"}
                res = _transform_group_result(_fv)
                return {_OUTPUT_JSON_FIELD: json.dumps({"result": res})}

            return fv_tool

    @classmethod
    def invoke(cls, agent: AgentBase, name: str, kwargs: Dict[str, Any]) -> Any:
        return agent.invoke_feature_view(name, kwargs)



import pandas as pd
import pyarrow as pa
from pydantic import Field
from pydantic.fields import FieldInfo
from tecton.types import (
    Array,
    Bool,
    Float64,
    Float32,
    Int32,
    Int64,
    Map,
    String,
    Timestamp,
    SdkDataType,
)
from tecton_core.data_types import DataType, ArrayType, MapType

import datetime
from typing import Annotated, Any, Tuple, get_args, get_origin, List, Dict, Union


def to_tecton_type(type: Any) -> Any:
    """
    Convert an object to a Tecton type

    Args:

        type: The Python type annotation

    Returns:

        Any: The Tecton type
    """
    if get_origin(type) is Annotated:
        type = get_args(type)[0]
    if isinstance(type, pa.DataType):
        if pa.types.is_string(type):
            return String
        if pa.types.is_int32(type):
            return Int32
        if pa.types.is_integer(type):
            return Int64
        if pa.types.is_floating(type):
            return Float64
        if pa.types.is_boolean(type):
            return Bool
        if pa.types.is_timestamp(type):
            return Timestamp
        if pa.types.is_list(type):
            return Array(to_tecton_type(type.value_type))
        raise ValueError(f"Unsupported type {type}")
    if type is str:
        return String
    if type is int:
        return Int64
    if type is float:
        return Float64
    if type is bool:
        return Bool
    if type is pd.Timestamp or type is datetime.datetime:
        return Timestamp
    if get_origin(type) is list:
        return Array(to_tecton_type(get_args(type)[0]))
    if get_origin(type) is dict:
        k = to_tecton_type(get_args(type)[0])
        v = to_tecton_type(get_args(type)[1])
        return Map(k, v)
    raise ValueError(f"Unsupported type {type}")


def tecton_type_to_python_annotation(dtype: Union[SdkDataType, DataType]) -> Any:
    tp = dtype.tecton_type if isinstance(dtype, SdkDataType) else dtype
    if String.tecton_type == tp:
        return str
    elif Int32.tecton_type == tp:
        return int
    elif Int64.tecton_type == tp:
        return int
    elif Float32.tecton_type == tp:
        return float
    elif Float64.tecton_type == tp:
        return float
    elif Bool.tecton_type == tp:
        return bool
    elif Timestamp.tecton_type == tp:
        return datetime.datetime
    elif isinstance(dtype, (Array, ArrayType)):
        return List[tecton_type_to_python_annotation(dtype.element_type)]
    elif isinstance(dtype, (Map, MapType)):
        return Dict[
            tecton_type_to_python_annotation(tp.key_type),
            tecton_type_to_python_annotation(tp.value_type),
        ]
    raise ValueError(f"Unsupported tecton type {type(tp)}")


def parse_python_annotation(annotation: Any) -> Tuple[Any, FieldInfo]:
    if get_origin(annotation) is Annotated:
        args = get_args(annotation)
        if len(args) != 2:
            raise ValueError(
                "Annotated should have exactly 2 arguments, type and description"
            )
        if isinstance(args[1], str):
            return args[0], Field(description=args[1])
        elif isinstance(args[1], FieldInfo):
            return args[0], args[1]
        else:
            raise ValueError("Invalid Annotated type")
    return annotation, Field()



import argparse
from contextlib import contextmanager
from typing import Any

from agents import build_agent

@contextmanager
def init_copilot(llm: Any):
    from melvin.tecton_gen_ai.api import Configs
    from melvin.tecton_gen_ai.testing import set_dev_mode

    set_dev_mode()
    with Configs(llm=llm, agent_invoke_kwargs={"max_iterations": 30}).update_default():
        yield build_agent()


def parse_args():
    parser = argparse.ArgumentParser(description="Q&A or chat with tecton docs.")
    # add a model parameter with short form -m to specify model
    parser.add_argument(
        "--model",
        "-m",
        type=str,
        default="openai/gpt-4o-2024-11-20",
        # default="openai/o3-mini-2025-01-31",
        help="LLM to use",
    )
    # add a temperature parameter with short form -t to specify temperature
    parser.add_argument(
        "--temperature",
        "-t",
        type=float,
        default=0.0,
        help="temperature for sampling",
    )
    # add a max_length parameter with short form -l to specify max_length
    parser.add_argument(
        "--max_tokens",
        "-l",
        type=int,
        default=4096,
        help="maximum token length of the output",
    )

    # parse the input
    args = parser.parse_args()

    llm = {
        "model": args.model,
        "temperature": args.temperature,
        "max_tokens": args.max_tokens,
    }
    res = vars(args)
    res["llm"] = llm
    return res



import logging
import io
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Iterator


class CostMonitor:
    def __init__(self):
        self.total_cost = 0.0
        self.total_requests = 0.0

    def report(self, cost: float, requests: float):
        self.total_cost += cost
        self.total_requests += requests


_COST_MONITOR = ContextVar("_COST_MONITOR", default=None)


@contextmanager
def cost_monitor() -> Iterator[CostMonitor]:
    try:
        monitor = CostMonitor()
        token_cost = _COST_MONITOR.set(monitor)
        yield monitor
    finally:
        _COST_MONITOR.reset(token_cost)


def get_cost_monitor() -> CostMonitor:
    res = _COST_MONITOR.get()
    if res is None:
        raise RuntimeError("Cost monitor is not set")
    return res


def report_cost(cost: float, requests: float) -> None:
    res = _COST_MONITOR.get()
    if res is not None:
        get_cost_monitor().report(cost, requests)


def make_noop_logger() -> logging.Logger:
    null_logger = logging.getLogger("NOOP")
    null_logger.handlers.clear()
    null_logger.addHandler(logging.NullHandler())  # read below for reason
    null_logger.propagate = False
    return null_logger


def make_string_logger(name: str, sio: io.StringIO) -> logging.Logger:
    string_logger = logging.getLogger(name)
    string_logger.handlers.clear()
    string_logger.addHandler(logging.StreamHandler(sio))
    string_logger.propagate = False
    return string_logger


NOOP_LOGGER = make_noop_logger()



import streamlit as st
from melvin.tecton_gen_ai.api import Agent
from utils import _log, load_object, store_object
from typing import Callable
from editors.feature_view_editor_ui import display_editor, FeatureEditorState

def _load_fv_editor_state(session_object_key: str) -> FeatureEditorState:
    try:
        state = load_object(session_object_key)
        state.session_object_key = session_object_key
        return state
    except Exception as e:
        _log(f"Error loading editor state: {e}")
        return FeatureEditorState(result_apply=False, result_cancel=False, session_object_key=session_object_key)

def _complete_session(editor_state: FeatureEditorState):
    st.session_state[editor_state.session_object_key] = True
    store_object(editor_state, "editor", editor_state.session_object_key)

def _create_fv_from_state(editor_state: FeatureEditorState, simulate_user_prompt_fn: Callable):
    submit_user_prompt = f"Create the following feature views for me. Create a new data source it doesn't exist. Create a new entity if it doesn't exist. Make up the FV name: {editor_state.to_dict()}"
    simulate_user_prompt_fn(submit_user_prompt)

def _on_apply(editor_state: FeatureEditorState, simulate_user_prompt_fn: Callable):
    editor_state.result_apply = True
    _complete_session(editor_state)
    _create_fv_from_state(editor_state, simulate_user_prompt_fn)

def _on_cancel(editor_state: FeatureEditorState):
    editor_state.result_cancel = True
    _complete_session(editor_state)


def _display_completed_session(editor_state: FeatureEditorState):
    st.code(f"Session completed: {editor_state.session_object_key}. Result: {editor_state.result_apply} {editor_state.result_cancel}")

def display_fv_editor(session_object_key: str, simulate_user_prompt_fn: Callable):
    editor_state = None

    # Load the editor state from the streamlit session state or from file if it was just created by the agent
    if f"{session_object_key}_ui_state" in st.session_state:
        editor_state = st.session_state[f"{session_object_key}_ui_state"]
    else:
        editor_state = _load_fv_editor_state(session_object_key)
        st.session_state[f"{session_object_key}_ui_state"] = editor_state

    if session_object_key in st.session_state:
        _display_completed_session(editor_state)
    else:
        x, y = st.columns(2)
        with x:
            st.code(session_object_key)
            display_editor(editor_state, key_prefix=session_object_key, on_ship_it=lambda: _on_apply(editor_state, simulate_user_prompt_fn))


def sys_prompt() -> str:
    return """
You are going to generate a UI-based feature view editor.

You will call one and only one editor tool based on the user's intent to output the feature view editor. If there's no clear user intent, make one up.

If not successful, you should return an error message starting with `Error: `
If successful, return a editor state representation which is a session object key starting with `so_editor_`,
and surround it with triple backticks, language is `tecton_editor`

The editor state must be generated by one of the tools. don't manipulate the `so_editor_` id.

Example of the editor state representation:

```tecton_editor
so_editor_sdgg
```
"""

def _generate_aggregation_feature_view_editor_state() -> str:
    """
    Generate aggregation feature view editor

        Returns:

            str: the editor state representation session object key, starting with `so_editor_`
    """

    editor_state = FeatureEditorState(
        result_apply=False,
        result_cancel=False,
    )
    return store_object(editor_state, "editor")


def make_fv_editor_agent() -> Agent:
    return Agent(
        name="editor_agent",
        description="""An agent that generates a graphical Feature View editor state

Returns:

    if not successful, it returns an error message starting with `Error: `
    if successful, it returns a Feature View editor state quoted with triple backticks, language is `tecton_editor`. You must show the editor state quoted in triple backticks to the user.

Note:

    - The output of this agent should never be used to call this agent again
    - For any follow up question, the input data needs to be regenerated
    - Do not try to interpret the content in a tecton_editor code block, use the exact code block in the final output.
""",
        prompt=sys_prompt,
        tools=[
            _generate_aggregation_feature_view_editor_state
        ],
    )



import json
import logging
import os
import shutil
from tempfile import gettempdir
from typing import Any, Dict, Optional, Union

from ..utils._dev_utils import set_dev_flag


def set_dev_mode():
    """
    Set the dev mode for testing and local development. In dev mode, you
    don't need to connect to tecton.
    """
    set_dev_flag()
    os.environ["TECTON_SKIP_OBJECT_VALIDATION"] = "true"
    os.environ["TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE"] = "rift"
    os.environ["TECTON_BATCH_COMPUTE_MODE"] = "rift"
    os.environ["TECTON_FORCE_FUNCTION_SERIALIZATION"] = "false"
    os.environ["DUCKDB_EXTENSION_REPO"] = ""


def make_local_vector_db_config(
    path: Optional[str] = None, remove_if_exists: bool = True
) -> Dict[str, Any]:
    """
    Create a local vector db configuration for testing.

    Args:

        path: The path to the local vector db, defaults to None (system temp directory / test.db).
        remove_if_exists: Whether to remove the existing db file, defaults to True.

    Returns:

        Dict[str, Any]: The local vector db configuration

    Note:

        - The local vector db is a LanceDB with OpenAIEmbeddings, it requires Langchain to be installed.
        - You need an OpenAI API key to use OpenAIEmbeddings.
        - Specifying the path is recommended to avoid conflicts with other tests.
    """
    if path is None:
        _path = os.path.join(gettempdir(), "tecton_local_vector_db_for_test.db")
    else:
        _path = path
    if remove_if_exists:
        shutil.rmtree(_path, ignore_errors=True)

    return {
        "provider": "lancedb",
        "embedding": "openai/text-embedding-3-small",
        "uri": _path,
    }


def make_debug_logger(filter: Any = None) -> logging.Logger:
    """
    Create a logger for debugging.

    Args:

        filter: The filter functionto apply to the logger, defaults to None (no filter).

    Returns:

        logging.Logger: The logger

    Example:

        ```python
        from tecton_gen_ai.testing.utils import make_debug_logger

        logger = make_debug_logger(filter=lambda record: "invoking" in record.getMessage())
        ```
    """
    from rich.logging import RichHandler

    logger = logging.getLogger("rich_logger")
    logger.setLevel(logging.DEBUG)
    handler = RichHandler(show_time=False)
    logger.handlers.clear()
    logger.addHandler(handler)
    if filter is not None:
        logger.filters.clear()
        logger.addFilter(filter)
    return logger


def print_md(text: Union[dict[str, Any], str]) -> None:
    """
    Print markdown text in a rich format if rich is installed, otherwise use IPython.

    Args:

        text: The markdown text to print or a python dict
    """
    if isinstance(text, dict):
        md = json.dumps(text, indent=2)
    else:
        md = None
    try:
        import rich
        from rich.markdown import Markdown

        rich.print(md or Markdown(text))
    except ImportError:
        from IPython.display import Markdown, display

        display(md or Markdown(text))



from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any, Dict, Iterator

import pandas as pd
from pydantic import BaseModel, Field

from ..constants import DEFAULT_TIMEOUT


class Configs(BaseModel):
    """
    The base configs for the feature view, LLM model, and feature service
    """

    llm: Any = Field(description="The LLM model", default=None)
    default_timeout: str = Field(
        description="The default timeout", default=DEFAULT_TIMEOUT
    )
    base_config: Dict[str, Any] = Field(
        description="The base config for all tecton objects", default={}
    )
    fv_base_config: Dict[str, Any] = Field(
        description="The base config for all types of feature views", default={}
    )
    bfv_config: Dict[str, Any] = Field(
        description="The config for the batch feature view", default={}
    )
    sfv_config: Dict[str, Any] = Field(
        description="The config for the stream feature view", default={}
    )
    rtfv_config: Dict[str, Any] = Field(
        description="The config for the realtime feature view", default={}
    )
    feature_service_config: Dict[str, Any] = Field(
        description="The config for the feature service", default={}
    )
    agent_invoke_kwargs: Dict[str, Any] = Field(
        description="The kwargs for the agent invoke", default={}
    )

    def update(self, other: "Configs") -> "Configs":
        """
        Update the configs with the other configs

        Args:

            other: The other configs

        Returns:

                The updated configs
        """
        return Configs(
            llm=other.llm or self.llm,
            default_timeout=other.default_timeout or self.default_timeout,
            base_config=self.base_config | other.base_config,
            fv_base_config=self.fv_base_config | other.fv_base_config,
            bfv_config=self.bfv_config | other.bfv_config,
            sfv_config=self.sfv_config | other.sfv_config,
            rtfv_config=self.rtfv_config | other.rtfv_config,
            feature_service_config=self.feature_service_config
            | other.feature_service_config,
            agent_invoke_kwargs=self.agent_invoke_kwargs | other.agent_invoke_kwargs,
        )

    def get_bfv_config(self, *configs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the config for the batch feature view merged with the base config

        Args:

            configs: The additional configs to merge

        Returns:

            The config for the batch feature view
        """
        return self.merge(
            self.base_config, self.fv_base_config, self.bfv_config, *configs
        )

    def get_rtfv_config(self, *configs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the config for the realtime feature view merged with the base config

        Args:

            configs: The additional configs to merge

        Returns:

            The config for the realtime feature view
        """
        return self.merge(
            self.base_config, self.fv_base_config, self.rtfv_config, *configs
        )

    def get_sfv_config(self, *configs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the config for the stream feature view merged with the base config

        Args:

            configs: The additional configs to merge

        Returns:

            The config for the stream feature view
        """
        return self.merge(
            self.base_config, self.fv_base_config, self.sfv_config, *configs
        )

    def get_agent_invoke_kwargs(self, *configs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get the kwargs for the agent invoke merged with the base config

        Args:

            configs: The additional configs to merge

        Returns:

            The kwargs for the agent invoke
        """
        return self.merge(self.base_config, self.agent_invoke_kwargs, *configs)

    def get_timeout_sec(self, timeout: Any) -> float:
        """
        Get the timeout in seconds

        Args:

            timeout: The timeout, if None, it will use the default timeout

        Returns:

            The timeout in seconds
        """
        return pd.to_timedelta(timeout or self.default_timeout).total_seconds()

    def set_default(self) -> None:
        """
        Set this config set to be the the default for the feature view,
        LLM model, and feature service

        Examples:

            ```python
            from tecton_gen_ai.api import Configs

            Configs(llm="openai/gpt-4o").set_default()
            ```
        """
        _TECTON_GEN_AI_CONFIG.set(self)

    @contextmanager
    def update_default(self) -> Iterator["Configs"]:
        """
        Update the default configs for the feature view, LLM model, and feature service

        Examples:

            ```python
            from tecton_gen_ai.api import Configs

            Configs(llm="openai/gpt-4o").set_default()

            with Configs(fv_base_config={"name": "my_fv"}).update_default():
                conf = Configs.get_default()
                assert conf.fv_base_config["name"] == "my_fv"
                assert conf.llm == "openai/gpt-4o"

            conf = Configs.get_default()
            assert len(conf.fv_base_config)==0
            ```
        """
        old_config = _TECTON_GEN_AI_CONFIG.get()
        try:
            token = _TECTON_GEN_AI_CONFIG.set(old_config.update(self))
            yield
        finally:
            _TECTON_GEN_AI_CONFIG.reset(token)

    @staticmethod
    def get_default() -> "Configs":
        """
        Get the current default configs for the feature view, LLM model, and feature service

        Returns:

            The default configs

        Examples:

            Examples:

            ```python
            from tecton_gen_ai.api import Configs

            Configs(llm="openai/gpt-4o").set_default()

            with Configs(fv_base_config={"name": "my_fv"}).update_default():
                conf = Configs.get_default()
                assert conf.fv_base_config["name"] == "my_fv"
                assert conf.llm == "openai/gpt-4o"

            conf = Configs.get_default()
            assert len(conf.fv_base_config)==0
            ```
        """
        return _TECTON_GEN_AI_CONFIG.get()

    def merge(self, *dicts: Dict[str, Any]) -> Dict[str, Any]:
        res: Dict[str, Any] = {}
        secrets: Dict[str, Any] = {}
        resources: Dict[str, Any] = {}
        for d in dicts:
            secrets.update(d.get("secrects", {}))
            resources.update(d.get("resource_providers", {}))
            res.update(d)
        if len(secrets) > 0:
            res["secrets"] = secrets
        if len(resources) > 0:
            res["resource_providers"] = resources
        return res


_TECTON_GEN_AI_CONFIG = ContextVar("tecton_gen_ai_config", default=Configs())



from melvin.tecton_gen_ai.testing import make_local_realtime_feature_view
import pandas as pd
import pytest


def test_mock_rtfv(tecton_unit_test):
    fv = make_local_realtime_feature_view(
        "fv", {"user_id": "user1", "age": 30}, ["user_id"]
    )

    events = pd.DataFrame([{"user_id": "user1"}])

    odf = fv.get_features_for_events(events).to_pandas()
    assert odf["fv__age"].iloc[0] == 30

    events = pd.DataFrame([{"user_id": "user2"}])

    with pytest.raises(Exception):
        odf = fv.get_features_for_events(events).to_pandas()



import json
from functools import lru_cache
from time import time

# for executing function declarations
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union  # noqa

from httpx import Client as HttpClient

from ..constants import _INPUT_JSON_FIELD, _OUTPUT_JSON_FIELD
from ..utils._serialization import deserialize_pydantic
from ..utils.config_wrapper import from_json_config, to_json_config
from ..utils.runtime import RuntimeVar, get_runtime_context
from .base import AgentBase, AgentInputModel, FeatureServiceConfig, ToolWrapper


@lru_cache
def _get_tecton_client(
    url: Any, api_key: Any, workspace: Any, timeout_sec: float
) -> Any:
    from tecton_client import TectonClient

    http_client = HttpClient(timeout=timeout_sec)
    return TectonClient(
        url=url, api_key=api_key, default_workspace_name=workspace, client=http_client
    )


def get_agent(
    name: str,
    url: Optional[str] = None,
    workspace: Optional[str] = None,
    api_key: Union[str, RuntimeVar, None] = None,
    metastore_update_interval: Any = "60s",
) -> AgentBase:
    """
    Create a client connecting to a deployed agent service.

    Args:

        name: The agent name
        url: The Tecton URL, defaults to None (will use the default tecton url from the environment)
        workspace: The workspace name, defaults to None (will use the environment variable TECTON_WORKSPACE)
        api_key: The API key defaults to None (will use the environment variable TECTON_API_KEY)
        metastore_update_interval: The interval expression to update the metastore, defaults to "60s"

    Returns:

        Agent: The client agent
    """
    return _AgentClientRemote(
        name=name,
        url=url,
        workspace=workspace,
        api_key=api_key,
        metastore_update_interval=metastore_update_interval,
    )


class _AgentClientRemote(AgentBase):
    def __init__(
        self,
        name: str,
        url: Optional[str] = None,
        workspace: Optional[str] = None,
        api_key: Union[str, RuntimeVar, None] = None,
        metastore_update_interval: Any = "60s",
    ):
        import pandas as pd

        super().__init__(name=name)

        self.service = FeatureServiceConfig(
            url=url,
            workspace=workspace,
            api_key=api_key,
            service=name,
        )
        self.interval = pd.to_timedelta(metastore_update_interval).total_seconds()
        self.last_check = 0
        self._static_metastore = None
        self._orig_key = api_key

    def __json_config_dict__(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "url": self.service.url,
            "workspace": self.service.workspace,
            "api_key": self._orig_key,
            "metastore_update_interval": self.interval,
        }

    @property
    def metastore(self) -> Dict[str, Any]:
        current = time()
        if self._static_metastore is None or current - self.last_check > self.interval:
            self._static_metastore = self._invoke(
                self.name + "_metastore",
                [],
                [],
                {},
                feature_type="metastore",
                timeout_sec=5,
            )
            self.last_check = current
        return self._static_metastore

    @property
    def input_schema(self) -> Type[AgentInputModel]:
        return deserialize_pydantic(self.metastore["input_schema"])

    @property
    def output_schema(self) -> Union[Type[str], Type[AgentInputModel]]:
        schema = self.metastore["output_schema"]
        if schema == "str":
            return str
        return deserialize_pydantic(schema)

    def export_tools(self, names: Optional[list[str]] = None) -> List[ToolWrapper]:
        tools = []
        for name in self.metastore.get("tools", {}).keys():
            if names is None or name in names:
                tools.append(_RemoteTool(self, name))
        return tools

    def _get_tecton_client(self, timeout_sec: float) -> Any:
        return _get_tecton_client(
            url=self.service.url,
            api_key=self.service.api_key,
            workspace=self.service.workspace,
            timeout_sec=timeout_sec,
        )

    def _invoke_entrypoint(
        self,
        message: str,
        timeout_sec: float,
        chat_history: Optional[List[Tuple[str, str]]] = None,
        context: Optional[Dict[str, Any]] = None,
        agent_name: Optional[str] = None,
    ) -> Union[str, Dict[str, Any]]:
        target_name = agent_name or self.name
        ctx = {
            "query": message,
            "chat_history": [list(x) for x in (chat_history or [])],
        }
        if context:
            ctx.update(context)
        input_json = json.dumps(ctx)
        session_path = get_runtime_context().get("session_path")
        client = self._get_tecton_client(timeout_sec=timeout_sec)
        gf = client.get_features(
            feature_service_name=target_name,
            request_context_map={
                _INPUT_JSON_FIELD: input_json,
                "session_path": session_path,
            },
        )
        fd = gf.get_features_dict()
        key = target_name + "." + _OUTPUT_JSON_FIELD
        resp = json.loads(fd[key])
        if "error" in resp:
            raise Exception(resp["error"])
        return resp["result"]

    def _get_feature_value(
        self,
        name: str,
        key_map: Dict[str, Any],
        request_map: Dict[str, Any],
        feature_type: str,
        timeout_sec: float,
    ):
        if feature_type == "prompt_not_in_use":  # pragma: no cover
            # TODO: prompt FCO: fix or delete?
            request_context_map = {
                "name": name,
                **request_map,
            }
        else:
            request_context_map = {
                "name": name,
                "input": json.dumps(request_map),
                "session_path": get_runtime_context().get("session_path"),
            }
        request_context_map["_dummy_request_ts"] = None
        client = self._get_tecton_client(timeout_sec=timeout_sec)
        gf = client.get_features(
            feature_service_name=self.name + "_" + name,
            join_key_map=key_map,
            request_context_map=request_context_map,
        )
        fd = gf.get_features_dict()
        if feature_type == "prompt_not_in_use":  # pragma: no cover
            # TODO: prompt FCO: fix or delete?
            return fd[name + ".prompt"]
        else:
            resp = json.loads(fd[name + "." + _OUTPUT_JSON_FIELD])
            if "error" in resp:
                raise Exception(resp["error"])
            result = resp["result"]
            return result


class _RemoteTool(ToolWrapper):
    subtype: str = "remote_tool"

    def __init__(
        self,
        agent: _AgentClientRemote,
        tool_name: str,
    ):
        self.agent = agent
        self._tool_name = tool_name
        self.metadata = agent.metastore["tools"][tool_name]

    @property
    def use_when(self) -> Optional[Callable]:
        return None

    @property
    def metastore_key(self) -> str:
        return self.tool_name

    @property
    def tool_name(self) -> str:
        return self.agent.name + "_" + self._tool_name

    def make_metadata(self):
        meta = self.metadata.copy()
        meta["agent_json"] = to_json_config(self.agent)
        meta["remote_name"] = self._tool_name
        base = self.make_base_metadata()
        base.pop("use_when")
        meta.update(base)
        return meta

    @classmethod
    def invoke(cls, agent: AgentBase, name: str, kwargs: Dict[str, Any]) -> Any:
        meta = agent.metastore["tools"][name]
        ctx = (
            agent._get_context()
        )  # TODO: should we pass the context to the remote agent?
        ctx.update(kwargs)
        _agent = from_json_config(meta["agent_json"])
        return _agent.invoke_tool(
            name=meta["remote_name"],
            kwargs=ctx,
        )



from melvin.tecton_gen_ai.utils.io_utils import (
    safe_write,
    safe_read,
    create_temp_dir,
    write_object,
    read_object,
)
from pytest import raises
import os


def test_safe_read_write(tmp_path):
    uri = str(tmp_path / "test1")
    with raises(FileNotFoundError):
        with safe_read(uri) as f:
            pass
    safe_write(uri, lambda f: f.write(b"hello"))
    with safe_read(uri) as f:
        assert f.read() == b"hello"
    safe_write(uri, lambda f: f.write(b"hello2"))  # no op since file exists
    with safe_read(uri) as f:
        assert f.read() == b"hello"
    uri = str(tmp_path / "test2")
    with open(uri, "w") as f:
        f.write("hello")
    with raises(FileExistsError):
        safe_write(uri, lambda f: f.write(b"hello"))
    path = tmp_path / "test3"
    # create a directory
    path.mkdir()
    path1 = path / "a"
    with open(str(path1), "w") as f:
        f.write("hello")
    path2 = path / "b"
    with open(str(path2), "w") as f:
        f.write("hello2")
    with raises(FileNotFoundError):
        with safe_read(str(path)) as f:
            pass
    path3 = path / "b.done"
    path3.touch()
    with safe_read(str(path)) as f:
        assert f.read() == b"hello2"


def test_create_temp_dir(tmp_path):
    with create_temp_dir(str(tmp_path), 1) as dir_path:
        path = os.path.join(dir_path, "a")
        with open(path, "w") as f:
            f.write("hello")
        with create_temp_dir(str(tmp_path), 1) as dir_path1:
            assert dir_path == dir_path1
            assert os.path.exists(dir_path)
        assert os.path.exists(dir_path)

    assert not os.path.exists(dir_path)


def test_read_write_object(tmp_path):
    with create_temp_dir(str(tmp_path), 1) as dir_path:
        write_object(1, "hello", dir_path)
        assert read_object(1, dir_path) == "hello"
        write_object(1, "hello2", dir_path)  # not overwritable
        assert read_object(1, dir_path) == "hello"
        write_object(2, "hello2", dir_path)
        assert read_object(2, dir_path) == "hello2"
        with raises(FileNotFoundError):
            read_object(3, dir_path)



import asyncio
import threading
from typing import Any, Coroutine, Generic, Optional, TypeVar, Union

T = TypeVar("T")


class _Sentinel: ...


class _CoroutineThread(threading.Thread, Generic[T]):
    """_CoroutineThread is a thread designed for running coroutines.

    It is cancelable and will propagate the cancel to the underlying coroutine.
    The result is stored in the `result` attribute.
    """

    def __init__(self, coroutine: Coroutine[Any, Any, T]):
        threading.Thread.__init__(self)
        self._coroutine = coroutine
        self._asyncio_loop = None
        self.asyncio_loop: Optional[asyncio.AbstractEventLoop] = None
        self._result: Union[T, _Sentinel] = _Sentinel()
        self.exception: Optional[Exception] = None

    @property
    def result(self) -> T:
        # We use a Sentinel object instead of `None`, since `None` could be a valid
        # output to a coroutine. This allows us to differentiate between the scenario where
        # we didn't complete execution, from where there's a valid output.
        if isinstance(self._result, _Sentinel):
            raise ValueError("Result not set")
        return self._result

    def run(self):
        self._asyncio_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._asyncio_loop)
        try:
            self._result = self._asyncio_loop.run_until_complete(self._coroutine)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.exception = e
        finally:
            self._asyncio_loop.close()

    def cancel(self):
        if self._asyncio_loop is not None:
            self._asyncio_loop.call_soon_threadsafe(self._cancel_coroutine)

    def _cancel_coroutine(self):
        tasks = asyncio.all_tasks(loop=self._asyncio_loop)
        for task in tasks:
            task.cancel()


def _run_coroutine_as_thread(coroutine: Coroutine[Any, Any, T]) -> T:
    """Runs a coroutine in a separate thread and returns the result."""
    thread = _CoroutineThread(coroutine)
    thread.start()

    try:
        thread.join()
        if thread.exception is not None:
            raise thread.exception
        return thread.result
    except KeyboardInterrupt:
        print("\nKeyboard interrupt received. Stopping the thread...")
        thread.cancel()
        thread.join()
        raise


def _is_running_loop() -> bool:
    """Utility that returns a boolean representing whether an an event loop is running"""
    try:
        return asyncio.get_running_loop() is not None
    except RuntimeError:
        return False


def run(coroutine: Coroutine[Any, Any, T]) -> T:
    """Runs a coroutine synchronously.

    If executing without an event loop (typical setting), it uses `asyncio.run`.
    If executing with an existing event loop (happens in Jupyter), it runs the coroutine in a separate thread.
    """
    if _is_running_loop():
        return _run_coroutine_as_thread(coroutine)
    else:
        return asyncio.run(coroutine)



import os
import pickle
import uuid
from datetime import date, datetime
from functools import wraps
from typing import Any, Callable, Optional, TypeVar

import pandas as pd
import pyarrow.fs as pfs
from typing_extensions import ParamSpec

from ..constants import _TECTON_CHECKPOINT_ATTR
from ..utils.hashing import to_uuid

from ._tecton_utils import get_cache_base_path

T = TypeVar("T")
P = ParamSpec("P")


def make_deterministic(obj: Any, key_obj: Any) -> Any:
    """Make an object deterministic by adding a checkpoint attribute to it.
    It's a no-op for primitive types and datetimes.

    Args:

        obj: The object to make idempotent
        key_obj: The key object to generate the determinitic key for the object

    Returns:

        The object with the checkpoint
    """
    if obj is None:
        return obj
    if isinstance(obj, (int, float, bool, str, date, datetime)):
        return obj
    if isinstance(obj, pd.DataFrame):
        setattr(obj, _TECTON_CHECKPOINT_ATTR, to_uuid(key_obj))
        return obj
    raise NotImplementedError(f"Type {type(obj)} is not supported")


def checkpoint(
    func: Callable[..., T],
    *args: Any,
    idempotency_key: Any = None,
    **kwargs: Any,
) -> T:
    """Checkpoint a function call. If the function has already been
    called with the same arguments, it will return the result from the cache.

    Args:

        func: The function to checkpoint
        args: The arguments to the function
        kwargs: The keyword arguments to the function
        idempotency_key: The extra key to change the function's idempotency, defaults to None

    Returns:

        The result of the function
    """
    uid = _func_to_uuid(func, *args, idempotency_key=idempotency_key, **kwargs)
    fs, base_path = get_cache_base_path()
    file_folder = os.path.join(base_path, uid + ".parquet")
    selector = pfs.FileSelector(
        file_folder,
        recursive=True,
        allow_not_found=True,
    )
    result = None
    for task in fs.get_file_info(selector):
        if task.path.endswith("/SUCCESS"):
            with fs.open_input_file(task.path) as f:
                data_type = pickle.load(f)
            file_path = os.path.join(
                task.path.rsplit("/", 1)[0],
                "data",
            )
            if issubclass(data_type, pd.DataFrame):
                result = pd.read_parquet(file_path, engine="pyarrow", filesystem=fs)
            else:
                with fs.open_input_file(file_path) as f:
                    result = pickle.load(f)
            break
    if result is None:
        task_folder = os.path.join(file_folder, "task_" + str(uuid.uuid4()))
        file_path = os.path.join(task_folder, "data")
        success_path = os.path.join(task_folder, "SUCCESS")
        result = func(*args, **kwargs)
        fs.create_dir(task_folder, recursive=True)
        if isinstance(result, pd.DataFrame):
            result.to_parquet(file_path, engine="pyarrow", filesystem=fs)
        else:
            with fs.open_output_stream(file_path) as f:
                pickle.dump(result, f)
        with fs.open_output_stream(success_path) as f:
            pickle.dump(type(result), f)
    return make_deterministic(result, uid)


def with_checkpoint(
    func: Optional[Callable[P, T]] = None,
    idempotency_key: Any = None,
) -> Callable[P, T]:
    """Decorator to checkpoint a function call. If the function has already been
    called with the same arguments, it will return the result from the cache.

    Args:

        func: The function to checkpoint
        idempotency_key: The extra key to change the function's idempotency, defaults to None

    Example:

        ```python
        @with_checkpoint
        def func1(a: int, b: str) -> pd.DataFrame:
            return random.random()

        @with_checkpoint(idempotency_key=2)
        def func2(df: pd.DataFrame) -> int:
            return 1
        ```
    """

    def _wrap(_func: Callable[P, T]) -> Callable[P, T]:
        @wraps(_func)
        def run(*args: P.args, **kwargs: P.kwargs) -> T:
            return checkpoint(
                _func,
                *args,
                idempotency_key=idempotency_key,
                **kwargs,
            )

        return run

    if func is not None:
        return _wrap(func)
    return _wrap


def _func_to_uuid(
    func: Callable[P, T],
    *args: Any,
    idempotency_key: Any = None,
    **kwargs: Any,
) -> str:
    return to_uuid(
        (
            func.__module__,
            func.__qualname__,
            args,
            kwargs,
            idempotency_key,
        )
    )



import asyncio
from typing import Any, Callable, TypeVar, Union

from pydantic import BaseModel

from . import _cache, _coroutine
from .hashing import to_uuid
from ._serialization import to_openai_function, openai_function_to_pydantic

_DEFAULT_LOCAL_CONCURRENCY = 5
T_OutputType = TypeVar("T_OutputType", bound=BaseModel)


def batch_generate_dicts(
    model: Union[str, dict[str, Any]],
    texts: list[str],
    schema: Union[type[T_OutputType], dict[str, Any]],
    concurrency: int = _DEFAULT_LOCAL_CONCURRENCY,
    enable_cache: bool = True,
) -> list[dict[str, Any]]:
    """
    Generate structured outputs for a list of texts

    Args:

        model: The model to use for generation
        texts: The list of texts to generate structured outputs for
        schema: The schema of the structured outputs
        concurrency: The number of concurrent requests to make, defaults to 5
        enable_cache: Whether to enable caching, defaults to True

    Returns:

        list[dict[str, Any]]: The structured outputs (python dicts)

    Examples:

        ```python
        from tecton_gen_ai.utils.structured_outputs import batch_generate_dicts
        from pydantic import BaseModel, Field

        model = "openai/gpt-4o"
        texts = ["I am 4", "She is 5"]

        class Age(BaseModel):
            age: int = Field(description="The age")

        outputs = batch_generate_dicts(model, texts, Age)
        ```
    """
    processor = BatchProcessor.make(model=model, schema=schema)
    return processor.batch_process(
        texts,
        concurrency=concurrency,
        keep_pydantic=False,
        enable_cache=enable_cache,
    )


def batch_generate(
    model: Union[str, dict[str, Any]],
    texts: list[str],
    schema: type[T_OutputType],
    concurrency: int = _DEFAULT_LOCAL_CONCURRENCY,
    enable_cache: bool = True,
) -> list[T_OutputType]:
    """
    Generate structured outputs for a list of texts

    Args:

        model: The model to use for generation
        texts: The list of texts to generate structured outputs for
        schema: The pydantic schema of the structured outputs
        concurrency: The number of concurrent requests to make, defaults to 5
        enable_cache: Whether to enable caching, defaults to True

    Returns:

        list[T_OutputType]: The structured outputs

    Examples:

        ```python
        from tecton_gen_ai.utils.structured_outputs import batch_generate

        model = "openai/gpt-4o"
        texts = ["I am 4", "She is 5"]

        class Age(BaseModel):
            age: int = Field(description="The age")

        outputs = batch_generate(model, texts, Age)
        ```
    """
    return [
        schema.model_validate(x)
        for x in batch_generate_dicts(
            model=model,
            texts=texts,
            schema=schema,
            concurrency=concurrency,
            enable_cache=enable_cache,
        )
    ]


class BatchProcessor:
    _PROCESSORS: list[
        tuple[Callable[[dict[str, Any]], bool], type["BatchProcessor"]]
    ] = []

    @staticmethod
    def register(fn: Callable[[dict[str, Any]], bool]):
        def _register(cls):
            BatchProcessor._PROCESSORS.append((fn, cls))
            return cls

        return _register

    @staticmethod
    def make(
        model: Union[str, dict[str, Any]],
        schema: Union[type[T_OutputType], dict[str, Any]],
    ) -> "BatchProcessor":
        if isinstance(model, str):
            model = {"model": model}
        _cls = BatchProcessor._get_cls(model)
        return _cls(model, schema)

    @staticmethod
    def _get_cls(model: dict[str, Any]) -> type["BatchProcessor"]:
        _load_dependencies()
        for fn, cls in BatchProcessor._PROCESSORS:
            if fn(model):
                return cls
        raise ValueError(f"Model {model} not supported")

    def __init__(
        self, model: dict[str, Any], schema: Union[type[T_OutputType], dict[str, Any]]
    ):
        self.model = model
        if isinstance(schema, dict):
            self.schema_model = openai_function_to_pydantic(schema)
            self.schema_dict = schema
        else:
            self.schema_model = schema
            self.schema_dict = to_openai_function(schema)

    def __json_config_dict__(self) -> dict[str, Any]:
        return {"model": self.model, "schema": self.schema_dict}

    def get_tecton_fields(self, as_attributes: bool = False) -> list:
        raise NotImplementedError

    async def aprocess(
        self, prompt: Union[str, list[tuple[str, str]]], keep_pydantic: bool
    ) -> Union[dict[str, Any], BaseModel]:
        raise NotImplementedError

    def batch_process(
        self,
        prompts: list[Union[str, list[tuple[str, str]]]],
        concurrency: int = _DEFAULT_LOCAL_CONCURRENCY,
        keep_pydantic: bool = False,
        enable_cache: bool = True,
    ) -> list[Union[dict[str, Any], BaseModel]]:
        # NOTE: default size limit of 1GB (https://grantjenks.com/docs/diskcache/api.html#constants)
        cache = _cache.get_cache("tecton-gen-ai", "structured_outputs")
        if enable_cache:
            base_key = to_uuid((self.schema_dict, self.model, keep_pydantic))
        else:
            base_key = ""

        async def fn(sem, prompt: Any) -> T_OutputType:
            async with sem:
                if not enable_cache:
                    return await self.aprocess(prompt, keep_pydantic=keep_pydantic)
                key = to_uuid((prompt, base_key))
                if (cached := await cache.aget(key)) is not None:
                    # Deserialize from JSON based on the return type
                    return cached

                result = await self.aprocess(prompt, keep_pydantic=keep_pydantic)
                await cache.aset(key, result)
                return result

        async def _gather():
            sem = asyncio.Semaphore(concurrency)
            _jobs = [fn(sem, prompt) for prompt in prompts]
            return await asyncio.gather(*_jobs)

        return _coroutine.run(_gather())


def _load_dependencies():
    try:
        from ..integrations import instructor  # noqa
    except ImportError:
        pass



import ast
import inspect
from typing import Any
from textwrap import dedent


def get_declaration(obj: Any, entrypoint_only: bool) -> str:
    """
    Get the function or class declaration with docstring.

    Args:

        obj: The object
        entrypoint_only: Whether to get only the entrypoint of the object (either the class __init__ or function)

    Returns:

        str: The object declaration with docstring
    """
    if inspect.isclass(obj) or inspect.isfunction(obj):
        src = dedent(inspect.getsource(obj))
        tree = ast.parse(src)
        tree = _traverse(
            tree, init_only=inspect.isclass(obj) if entrypoint_only else False
        )
        return ast.unparse(tree)
    raise ValueError(f"{obj} is not a class or function")


def _traverse(node: Any, init_only: bool) -> Any:
    if isinstance(node, ast.FunctionDef):
        name = node.name
        if name == "__init__" or (not init_only and not name.startswith("_")):
            nb = node.body
            if (
                len(nb) > 0
                and isinstance(nb[0], ast.Expr)
                and isinstance(nb[0].value, ast.Constant)
            ):
                nb = [nb[0], ast.Pass()]
            else:
                nb = [ast.Pass()]
            node.body = nb
            return node
        else:
            return None
    elif hasattr(node, "body"):
        body = []
        for x in node.body:
            xx = _traverse(x, init_only=init_only)
            if xx is not None:
                body.append(xx)
        node.body = body
        return node
    else:
        return node



import click
from typing import Dict
from utils import _err, _log, run_command
from tecton.cli.cli import cli


def get_command_help(command, prefix="") -> Dict:
    """Recursively get help text for a command and its subcommands"""
    result = {
        "name": prefix + command.name if prefix else command.name,
        "help": command.help,
        "options": [],
    }

    # Get options
    ctx = click.Context(command)
    for param in command.get_params(ctx):
        opt_record = param.get_help_record(ctx)
        if opt_record:
            result["options"].append({"name": opt_record[0], "help": opt_record[1]})

    # Get subcommands if this is a group
    if isinstance(command, click.Group):
        result["commands"] = []
        for cmd_name, cmd in sorted(command.commands.items()):
            if not cmd.hidden:
                sub_prefix = (
                    f"{prefix}{command.name} " if prefix else f"{command.name} "
                )
                result["commands"].append(get_command_help(cmd, sub_prefix))

    return result


def tecton_cli_help() -> dict:
    """
    Query the tecton CLI to get a structured representation of all available commands and their options.
    Can be used to explore the CLI's functionality, including commands for checking user identity,
    managing workspaces, and cluster connectivity.

    Returns:
        dict: A nested dictionary containing the CLI command structure with the following keys:
            - name (str): The command name
            - help (str): The command's help text
            - options (list): List of dictionaries containing option names and help text
            - commands (list): List of nested command dictionaries (for subcommands)
    """
    return get_command_help(cli)


def tecton_cli_execute(command: str = "") -> str:
    """
    Execute a tecton cli command
    Use the tecton_cli_help tool to figure out which commands you have at your disposal

    Args:

        command: tecton command to execute, including any flags for that command. Do not prefix with the name of the cli, tecton

    Returns:

        str: The result of the command. May indicate success or failure
    """
    try:
        _log(f"Running tecton cli command {command}")

        code, out, err = run_command(f"tecton {command}")
        if code != 0:
            return _err(f"{err}\n\n{out}")
        return out
    except Exception as e:
        return _err(e)



import os
import re
import tempfile
import uuid
from dataclasses import dataclass
from typing import Any

import cloudpickle
import threading
from tecton._internals.utils import cluster_url

# 7-bit C1 ANSI sequences
ANSI_ESCAPE = re.compile(
    r"""
    \x1B  # ESC
    (?:   # 7-bit C1 Fe (except CSI)
        [@-Z\\-_]
    |     # or [ for CSI, followed by a control sequence
        \[
        [0-?]*  # Parameter bytes
        [ -/]*  # Intermediate bytes
        [@-~]   # Final byte
    )
""",
    re.VERBOSE,
)

_TECTON_BLOCK_REGEX = re.compile(r"```tecton_(.*?)```", re.DOTALL)
_TECTON_BLOCK_PLACEHOLDER = "((tecton_block))"


_CODE_CONTAINER = None

_LOCK = threading.Lock()
_TECTON_ACCOUNT_INFO = None


def set_code_container(code_container):
    global _CODE_CONTAINER
    _CODE_CONTAINER = code_container


def get_code_container():
    return _CODE_CONTAINER


def get_cwd():
    # use python to get the current working directory
    import os

    return os.getcwd()


def run_command(command):
    import subprocess

    process = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = process.communicate()
    out = out.decode("utf-8") if out else ""
    out = ANSI_ESCAPE.sub("", out.replace("\n", "<|__newline__|>")).replace(
        "<|__newline__|>", "\n"
    )
    err = err.decode("utf-8") if err else ""
    err = ANSI_ESCAPE.sub("", err.replace("\n", "<|__newline__|>")).replace(
        "<|__newline__|>", "\n"
    )
    return process.returncode, out, err


def validate_syntax(path: str) -> None:
    code, out, err = run_command(f"python -m py_compile {path}")
    if code != 0:
        raise SyntaxError(str(err))


def is_in_feature_repo():
    return _is_feature_repo(get_cwd())


def _is_feature_repo(path: str) -> bool:
    import os

    path = os.path.join(path, ".tecton")
    return os.path.exists(path)


def _log(msg: str) -> str:
    from melvin.tecton_gen_ai.testing.log_utils import get_ui_logger

    logger = get_ui_logger()
    extra = {"copilot_flag": "log"}
    logger.info(msg, extra=extra)
    return msg


def _log_with_details(msg: str, details: str) -> str:
    from melvin.tecton_gen_ai.testing.log_utils import get_ui_logger

    logger = get_ui_logger()
    extra = {"copilot_flag": "log_with_details", "details": details}
    logger.info(msg, extra=extra)
    return msg


def _log_image(title: str, image_base64: str) -> str:
    from melvin.tecton_gen_ai.testing.log_utils import get_ui_logger

    logger = get_ui_logger()
    extra = {"copilot_flag": "image", "title": title, "image_base64": image_base64}
    logger.info("Rendering image", extra=extra)
    return title


def _success(msg: str) -> str:
    from melvin.tecton_gen_ai.testing.log_utils import get_ui_logger

    logger = get_ui_logger()
    extra = {"copilot_flag": "success"}
    logger.info(msg, extra=extra)
    return msg


def _err(_msg: Any, prefix: str = "Error") -> str:
    from melvin.tecton_gen_ai.testing.log_utils import get_ui_logger

    msg = str(_msg)

    logger = get_ui_logger()
    extra = {"copilot_flag": "error"}
    logger.error(msg, extra=extra)
    return prefix + ": " + msg


def _code():
    container = get_code_container()
    if container is not None:
        container.code(_get_source_code(), language="python")


@dataclass
class TectonAccountInfo:
    """Represents system information about the tecton account. Cluster is sometimes also referred to as account or endpoint."""

    rift_compute_supported: bool  # Whether Rift Compute is supported
    spark_compute_supported: bool  # Whether Spark Compute is supported
    caller_identity: str
    cluster_url: str

    def to_dict(self) -> dict:
        return {
            "rift_compute_supported": self.rift_compute_supported,
            "spark_compute_supported": self.spark_compute_supported,
            "caller_identity": self.caller_identity,
            "cluster_url": self.cluster_url,
        }



def _get_current_tecton_account_info() -> dict:
    """
    Returns information about the currently logged in Tecton account.
    This information is important as it indicates what types of tecton capabilities are available or not supported.

    Returns:
        dict: A dictionary containing account information with the following keys:
            - rift_compute_supported (bool): Whether Rift Compute is supported
            - spark_compute_supported (bool): Whether Spark Streaming is supported
    """
    import tecton

    # Ensure we're logged into a cluster
    caller_identity = str(tecton.get_caller_identity())

    tecton.conf._force_initialize_mds_config()
    batch_compute_mode = tecton.conf._get("TECTON_BATCH_COMPUTE_MODE")
    print(tecton.conf._get("TECTON_BATCH_COMPUTE_MODE"))
    print(tecton.version.summary())

    rift_compute_supported = batch_compute_mode == "rift"
    spark_compute_supported = batch_compute_mode == "spark"

    return TectonAccountInfo(
        rift_compute_supported=rift_compute_supported,
        spark_compute_supported=spark_compute_supported,
        caller_identity=caller_identity,
        cluster_url=cluster_url(),
    ).to_dict()

def initialize_tecton_account_info_if_needed():
    global _TECTON_ACCOUNT_INFO
    if _TECTON_ACCOUNT_INFO is None:
        with _LOCK:
            _TECTON_ACCOUNT_INFO = _get_current_tecton_account_info()

def get_tecton_account_info() -> dict:
    # We cannot initialize lazily because of a race condition. ui.py sets dev_mode, which screws up the result
    with _LOCK:
        if _TECTON_ACCOUNT_INFO is None:
            raise Exception("Tecton account info not initialized")

        return _TECTON_ACCOUNT_INFO

def display(markdown_text, write_text_func, write_block_func):
    code_snippets = _TECTON_BLOCK_REGEX.findall(markdown_text)
    # Replace all found code snippets with <>
    replaced_text = _TECTON_BLOCK_REGEX.sub(_TECTON_BLOCK_PLACEHOLDER, markdown_text)
    n = 0
    parts = replaced_text.split(_TECTON_BLOCK_PLACEHOLDER)
    if len(parts) <= 1:
        write_text_func(markdown_text)
        return
    if parts[0]:
        write_text_func(parts[0])
    for part in parts[1:]:
        tp, code = code_snippets[n].split("\n", 1)
        write_block_func(tp, code)
        if part:
            write_text_func(part)
        n += 1


def _get_source_code() -> str:
    if not is_in_feature_repo():
        return ""

    path = os.path.join(get_cwd(), "features.py")
    if not os.path.exists(path):
        return ""
    else:
        with open(path, "r") as f:
            return f.read()


def store_object(obj, type_str: str, name: str = None) -> str:
    temp_dir = tempfile.gettempdir()
    _dir = os.path.join(temp_dir, "tecton_copilot")
    name = name or ("so_" + type_str + "_" + str(uuid.uuid4())[:5])
    path = os.path.join(_dir, name)

    os.makedirs(_dir, exist_ok=True)
    with open(path, "wb") as f:
        cloudpickle.dump(obj, f)
    # _log(f"Stored {name}")
    return name


def load_object(name: str) -> Any:
    temp_dir = tempfile.gettempdir()
    path = os.path.join(temp_dir, "tecton_copilot", name)
    with open(path, "rb") as f:
        return cloudpickle.load(f)



from typing import Tuple, Any, Dict

_ATTR_STORE: Dict[int, Tuple[Any, Dict[str, Any]]] = {}


def set_attr(obj: Any, key: str, value: Any) -> None:
    if id(obj) not in _ATTR_STORE:
        _ATTR_STORE[id(obj)] = (obj, {key: value})
    else:
        _ATTR_STORE[id(obj)][1][key] = value


def get_attr(obj: Any, key: str) -> Any:
    return _ATTR_STORE[id(obj)][1][key]


def has_attr(obj: Any, key: str) -> bool:
    if id(obj) not in _ATTR_STORE:
        return False
    return key in _ATTR_STORE[id(obj)][1]




from melvin.tecton_gen_ai.api import Agent

from tools import list_workspaces, list_feature_services, tecton_cli_help, tecton_cli_execute, get_feature_view_configuration, get_feature_service_configuration

DIAGONISTICS_KNOWLEDGE = """
# Batch Features

## Missing Data
Tecton silently ignores all rows that don’t fall into the materialization window.

### Common gotchas:
- Using `context.end_time` as the materialization run’s feature timestamp, and forgetting to subtract a millisecond.

## Inefficient Backfills
Not filtering the upstream data source(s) properly by the materialization run’s time window.

### Common gotchas:
- Failing to use `FilteredSource`.
- Failing to make the onboarded data source “filter-able”.

Using incremental backfills when you don’t need to.

### Common gotchas:
- Implemented a custom aggregation that can really be expressed using Tecton’s aggregation engine.
- Setting `incremental_backfill` to `True` when the transformation is a simple row-level transformation.

## Inefficient Provisioning
Not properly provisioning the compute infrastructure.

### Common gotchas:
- Provisioning too little memory.
- Provisioning too much memory.
- Provisioning not enough cores.

## Inefficient Compute Mode
Picking a suboptimal compute engine.

### Common gotchas:
- Using Python mode when Pandas should be used.
- Using Pandas mode when SparkSQL or SnowflakeSQL should be used.

---

# Streaming Features

## Inefficient Stream Processing
- Using continuous when `stream_tiling` should be used (to write less data to the online store).

## Staler-than-expected Streaming Features
- Using continuous Spark streaming features when continuous Rift features should be used.
- Using tiled streaming features when continuous should be used.
- Using `StreamIngest v1` instead of `StreamIngest v2`.

## Inefficient Provisioning
Under or over-provisioned resources.

## Inefficient Choice of Streaming Features
Using streaming features when batch processing is good enough.

### Common gotchas:
- Using streaming features for features that don’t decay or change quickly.

---

# Realtime Features

## Inefficient Compute
- Using Pandas mode when Python should be used.
- Using Python mode when Calculation features should be used.
- Using real-time computed features when features should be precomputed (e.g., using last-n + RTFV when you can just aggregate in a streaming feature).

## Inefficient Provisioning

### Common gotchas:
- `FeatureServer` (or `TSG`) is not properly statically provisioned.
- `FeatureServer` (or `TSG`) uses static provisioning but should use autoscaling.

---

# Training Data Generation

## Slow Training Data Generation

### Common gotchas:
- Failure to provision the job with enough memory.
- Failure to provision the job with enough nodes (if it can easily be parallelized).
- Failure to materialize features to the offline store.
- Having unnecessarily high TTL.
- Having unnecessarily large time windows.
- Using local mode when you should use remote dataset generation

---

# Real-Time Serving

## Expensive Online Store
Choosing Dynamo when Redis should be used.

### Common gotchas:
- Extremely write-heavy workloads.

Choosing Redis when Dynamo should be used.

### Common gotchas:
- Extremely low-read intense workloads.

## Expensive Online Store for Backfills

### Common gotchas:
- Failure to use bulk-import with Dynamo.

## High Read-Time Latencies / Expensive Online Compute

### Common gotchas:
- Time window is too long and compaction is disabled.
"""

def sys_prompt() -> str:
    return f"""
You are a helpful assistant that can help diagnose performance issues and answer questions about how to optimize performance.

Use the following knowledge base to help you answer the question. If this doesn't help, try to find more information in the documentation.
If the user is asking about a specific FeatureView or FeatureService, retrieve the object, inspect it and use it to make a better diagnosis.

{DIAGONISTICS_KNOWLEDGE}

The output should be a detailed explanation of the issue and the steps you can take to fix it. Include a link to relevant documentation whenever possible.
"""

def _make_agent(doc_retriever: Agent) -> Agent:
    return Agent(
        name="diagnostics_agent",
        description="""An agent that can help you diagnose performance issues and answer questions about how to optimize performance

Args:

    - query: A detailed summary of the issue the user is experiencing or the question they're asking. Include any previous context as relevant.

Returns:

    Returns an explanation of what the user can do to fix the issue or it may contain follow up questions

Note:

    - You must pass the agent response to the user. You must not call the agent multiple times.
""",
        prompt=sys_prompt,
        tools=[
            doc_retriever,
            list_workspaces,
            list_feature_services,
            get_feature_service_configuration,
            get_feature_view_configuration,
            tecton_cli_help,
            tecton_cli_execute,
        ],
    )



import os
from contextlib import contextmanager
from typing import Any, Callable, Iterator, Optional
from uuid import uuid4

import cloudpickle
from pyarrow.fs import FileSelector, FileSystem, FileType

from .hashing import to_uuid


def safe_write(uri: str, write_func: Callable):
    _SafePathObject(uri).safe_write(write_func)


@contextmanager
def safe_read(uri: str):
    obj = _SafePathObject(uri)
    with obj.safe_read() as f:
        yield f


@contextmanager
def create_temp_dir(base_uri: str, key: Any) -> Iterator[str]:
    dir_path = os.path.join(base_uri, to_uuid(key))
    fs, path = FileSystem.from_uri(dir_path)
    info = fs.get_file_info(path)
    existed = False
    if info.type == FileType.Directory:
        existed = True
    elif info.type != FileType.NotFound:
        raise FileExistsError(f"{dir_path} exists and it is not a directory")
    if not existed:
        fs.create_dir(path, recursive=True)
    yield dir_path
    if not existed:
        fs.delete_dir(path)


def write_object(key: Any, value: Any, base_uri: str) -> None:
    _key = to_uuid(key)
    path = os.path.join(base_uri, _key)
    safe_write(path, lambda f: cloudpickle.dump(value, f))


def read_object(key: Any, base_uri: str) -> Any:
    path = os.path.join(base_uri, to_uuid(key))
    with safe_read(path) as f:
        return cloudpickle.load(f)


class _SafePathObject:
    """
    A lockless thread safe file reader/writer
    """

    def __init__(self, uri: str):
        self.uri = uri
        self.fs, self.path = FileSystem.from_uri(uri)
        self.info = self.fs.get_file_info(self.path)
        if self.info.type == FileType.File or self.info.type == FileType.Unknown:
            raise FileExistsError(f"{uri} exists and it is not a directory")

    @contextmanager
    def safe_read(self):
        path = self._find_ready_path()
        if path is None:
            raise FileNotFoundError(f"{self.uri} is not ready")
        with self.fs.open_input_stream(path) as f:
            yield f

    def safe_write(self, write_func: Callable) -> None:
        if self._find_ready_path() is not None:
            return
        if self.info.type == FileType.NotFound:
            self.fs.create_dir(self.path, recursive=True)
        path = os.path.join(self.path, str(uuid4()))
        done = path + ".done"
        with self.fs.open_output_stream(path) as f:
            write_func(f)
        with self.fs.open_output_stream(done) as f:
            pass

    def _find_ready_path(self) -> Optional[str]:
        if self.info.type == FileType.NotFound:
            return None
        files = self.fs.get_file_info(FileSelector(self.path, recursive=False))
        for file in files:
            if file.path.endswith(".done"):
                return file.path[:-5]
        return None
